# Copyright (C) 2024-2026 Free Software Foundation, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

# This file exists to satisfy the standard layout of a Python program:
#   - a single .py file in the parent directory,
#   - any number of .py files in a subdirectory.
# Explanation: https://stackoverflow.com/questions/16981921/relative-imports-in-python-3
# Without this standard layout, we got
#   - errors at runtime:
#     from . import constants          =>   ImportError: attempted relative import with no known parent package
#     from pygnulib import constants   =>   ModuleNotFoundError: No module named 'pygnulib'
#   - error markers in Eclipse + PyDev.

from pygnulib import main

if __name__ == '__main__':
    main.main_with_exception_handling()



# Copyright (C) 2002-2026 Free Software Foundation, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

#===============================================================================
# Define global imports
#===============================================================================
import os
import re
import sys
import filecmp
import shlex
import subprocess as sp
from .constants import (
    DIRS,
    ensure_writable,
    hardlink,
    joinpath,
    link_if_changed,
    movefile,
    copyfile,
    substart,
)
from .enums import CopyAction
from .GLError import GLError
from .GLConfig import GLConfig


#===============================================================================
# Define GLFileSystem class
#===============================================================================
class GLFileSystem:
    '''GLFileSystem class is used to create virtual filesystem, which is based
    on the gnulib directory and directories specified by localpath argument.
    Its main method lookup(file) is used to find file in these directories or
    combine it using Linux 'patch' utility.'''

    config: GLConfig

    def __init__(self, config: GLConfig) -> None:
        '''Create new GLFileSystem instance. The only argument is localpath,
        which can be an empty list.'''
        if type(config) is not GLConfig:
            raise TypeError('config must be a GLConfig, not %s'
                            % type(config).__name__)
        self.config = config

    def __repr__(self) -> str:
        '''x.__repr__ <==> repr(x)'''
        result = '<pygnulib.GLFileSystem %s>' % hex(id(self))
        return result

    def lookup(self, name: str) -> tuple[str, bool]:
        '''Lookup a file in gnulib and localpath directories or combine it using
        'patch' utility. If file was found, method returns string, else it raises
        GLError telling that file was not found. Function also returns flag which
        indicates whether file is a temporary file.
        GLConfig: localpath.'''
        if type(name) is not str:
            raise TypeError('name must be a string, not %s'
                            % type(name).__name__)
        localpath = self.config['localpath']
        # Each element in localpath is a directory whose contents overrides
        # or amends the result of the lookup in the rest of localpath and
        # the gnulib dir. So, the first element of localpath is the highest
        # priority one.
        lookedupFile = None
        lookedupPatches = []
        for localdir in localpath:
            file_in_localdir = joinpath(localdir, name)
            if os.path.isfile(file_in_localdir):
                lookedupFile = file_in_localdir
                break
            diff_in_localdir = joinpath(localdir, '%s.diff' % name)
            if os.path.isfile(diff_in_localdir):
                lookedupPatches.append(diff_in_localdir)
        # Treat the gnulib dir like a lowest-priority --local-dir, except that
        # here we don't look for .diff files.
        if lookedupFile == None:
            file_in_localdir = joinpath(DIRS['root'], name)
            if os.path.isfile(file_in_localdir):
                lookedupFile = file_in_localdir
        if lookedupFile != None:
            if len(lookedupPatches) > 0:
                # Apply the patches, from lowest-priority to highest-priority.
                tempFile = joinpath(self.config['tempdir'], name)
                try:  # Try to create directories
                    os.makedirs(os.path.dirname(tempFile))
                except OSError:
                    pass  # Skip errors if directory exists
                if os.path.isfile(tempFile):
                    os.remove(tempFile)
                copyfile(lookedupFile, tempFile)
                ensure_writable(tempFile)
                for diff_in_localdir in reversed(lookedupPatches):
                    command = f'patch -s {shlex.quote(tempFile)} < {shlex.quote(diff_in_localdir)}'
                    try:  # Try to apply patch
                        sp.check_call(command, shell=True, stdout=sys.stderr)
                    except sp.CalledProcessError as exc:
                        raise GLError(2, name) from exc
                result = (tempFile, True)
            else:
                result = (lookedupFile, False)
        else:
            raise GLError(1, name)
        return result

    def shouldLink(self, original: str, lookedup: str) -> bool:
        '''GLFileSystem.shouldLink(original, lookedup)

        Determines whether the original file should be copied, symlinked,
          or hardlinked.
        Returns a CopyAction.'''
        copymode = self.config['copymode']
        lcopymode = self.config['lcopymode']
        localpath = self.config['localpath']

        # Don't bother checking the localpath's if we end up performing the same
        # action anyways.
        if copymode != lcopymode:
            for localdir in localpath:
                if lookedup == joinpath(localdir, original):
                    return lcopymode
        return copymode


#===============================================================================
# Define GLFileAssistant class
#===============================================================================
class GLFileAssistant:
    '''GLFileAssistant is used to help with file processing.'''

    original: str | None
    rewritten: str | None
    added: list[str]
    config: GLConfig
    transformers: dict[str, tuple[re.Pattern, str] | None]
    filesystem: GLFileSystem

    def __init__(self, config: GLConfig, transformers: dict[str, tuple[re.Pattern, str] | None] | None = None) -> None:
        '''Create GLFileAssistant instance.

        config stores information shared between classes.
        transformers is a dictionary which uses a file category as the key. The
          value accessed is a tuple containing arguments for re.sub() or None if
          no transformations are needed.'''
        if type(config) is not GLConfig:
            raise TypeError('config must be a GLConfig, not %s'
                            % type(config).__name__)
        if transformers == None:
            transformers = dict()
        elif type(transformers) is not dict:
            raise TypeError('transformers must be a dict, not %s'
                            % type(transformers).__name__)
        for key in ['lib', 'aux', 'main', 'tests']:
            if key not in transformers:
                transformers[key] = None
            else:  # if key in transformers
                value = transformers[key]
                if type(value) is not tuple and value != None:
                    raise TypeError('transformers[%s] must be a tuple or None, not %s'
                                    % (key, type(value).__name__))
        self.original = None
        self.rewritten = None
        self.added = []
        self.config = config
        self.transformers = transformers
        self.filesystem = GLFileSystem(self.config)

    def __repr__(self) -> str:
        '''x.__repr__() <==> repr(x)'''
        result = '<pygnulib.GLFileAssistant %s>' % hex(id(self))
        return result

    def tmpfilename(self, path: str) -> str:
        '''Return the name of a temporary file (file is relative to destdir).'''
        if type(path) is not str:
            raise TypeError('path must be a string, not %s'
                            % (type(path).__name__))
        if not self.config['dryrun']:
            # Put the new contents of $file in a file in the same directory (needed
            # to guarantee that an 'mv' to "$destdir/$file" works).
            result = joinpath(self.config['destdir'], '%s.tmp' % path)
            dirname = os.path.dirname(result)
            if dirname and not os.path.isdir(dirname):
                os.makedirs(dirname)
        else:  # if self.config['dryrun']
            # Put the new contents of $file in a file in a temporary directory
            # (because the directory of "$file" might not exist).
            tempdir = self.config['tempdir']
            result = joinpath(tempdir, '%s.tmp' % os.path.basename(path))
            dirname = os.path.dirname(result)
            if not os.path.isdir(dirname):
                os.makedirs(dirname)
        return result

    def setOriginal(self, original: str) -> None:
        '''Set the name of the original file which will be used.'''
        if type(original) is not str:
            raise TypeError('original must be a string, not %s'
                            % (type(original).__name__))
        self.original = original

    def setRewritten(self, rewritten: str) -> None:
        '''Set the name of the rewritten file which will be used.'''
        if type(rewritten) is not str:
            raise TypeError('rewritten must be a string, not %s'
                            % type(rewritten).__name__)
        self.rewritten = rewritten

    def addFile(self, file: str) -> None:
        '''Add file to the list of added files.'''
        if file not in self.added:
            self.added.append(file)

    def getFiles(self) -> list[str]:
        '''Return list of the added files.'''
        return list(self.added)

    def add(self, lookedup: str, tmpflag: bool, tmpfile: str) -> None:
        '''This method copies a file from gnulib into the destination directory.
        The destination is known to exist. If tmpflag is True, then lookedup file
        is a temporary one.'''
        original = self.original
        rewritten = self.rewritten
        destdir = self.config['destdir']
        if original == None:
            raise TypeError('original must be set before applying the method')
        if rewritten == None:
            raise TypeError('rewritten must be set before applying the method')
        if not self.config['dryrun']:
            print('Copying file %s' % rewritten)
            if self.filesystem.shouldLink(original, lookedup) == CopyAction.Symlink \
                    and not tmpflag and filecmp.cmp(lookedup, tmpfile):
                link_if_changed(lookedup, joinpath(destdir, rewritten))
            else:  # if any of these conditions is not met
                if self.filesystem.shouldLink(original, lookedup) == CopyAction.Hardlink \
                   and not tmpflag and filecmp.cmp(lookedup, tmpfile):
                    hardlink(lookedup, joinpath(destdir, rewritten))
                else:  # Move instead of linking.
                    try:  # Try to move file
                        movefile(tmpfile, joinpath(destdir, rewritten))
                    except Exception as exc:
                        raise GLError(17, original) from exc
        else:  # if self.config['dryrun']
            print('Copy file %s' % rewritten)

    def update(self, lookedup: str, tmpflag: bool, tmpfile: str, already_present: bool) -> None:
        '''This method copies a file from gnulib into the destination directory.
        The destination is known to exist. If tmpflag is True, then lookedup file
        is a temporary one.'''
        original = self.original
        rewritten = self.rewritten
        destdir = self.config['destdir']
        if original == None:
            raise TypeError('original must be set before applying the method')
        if rewritten == None:
            raise TypeError('rewritten must be set before applying the method')
        if type(lookedup) is not str:
            raise TypeError('lookedup must be a string, not %s'
                            % type(lookedup).__name__)
        if type(already_present) is not bool:
            raise TypeError('already_present must be a bool, not %s'
                            % type(already_present).__name__)
        basename = rewritten
        backupname = '%s~' % basename
        basepath = joinpath(destdir, basename)
        backuppath = joinpath(destdir, backupname)
        if not filecmp.cmp(basepath, tmpfile):
            if not self.config['dryrun']:
                if already_present:
                    print('Updating file %s (backup in %s)' % (basename, backupname))
                else:  # if not already_present
                    message = 'Replacing file '
                    message += '%s (non-gnulib code backed up in ' % basename
                    message += '%s) !!' % backupname
                    print(message)
                if os.path.isfile(backuppath):
                    os.remove(backuppath)
                try:  # Try to replace the given file
                    movefile(basepath, backuppath)
                except Exception as exc:
                    raise GLError(17, original) from exc
                if self.filesystem.shouldLink(original, lookedup) == CopyAction.Symlink \
                        and not tmpflag and filecmp.cmp(lookedup, tmpfile):
                    link_if_changed(lookedup, basepath)
                else:  # if any of these conditions is not met
                    if self.filesystem.shouldLink(original, lookedup) == CopyAction.Hardlink \
                       and not tmpflag and filecmp.cmp(lookedup, tmpfile):
                        hardlink(lookedup, basepath)
                    else:  # Move instead of linking.
                        try:  # Try to move file
                            if os.path.exists(basepath):
                                os.remove(basepath)
                            copyfile(tmpfile, joinpath(destdir, rewritten))
                        except Exception as exc:
                            raise GLError(17, original) from exc
            else:  # if self.config['dryrun']
                if already_present:
                    print('Update file %s (backup in %s)' % (rewritten, backupname))
                else:  # if not already_present
                    print('Replace file %s (backup in %s)' % (rewritten, backupname))

    def add_or_update(self, already_present: bool) -> None:
        '''This method handles a file that ought to be present afterwards.'''
        original = self.original
        rewritten = self.rewritten
        if original == None:
            raise TypeError('original must be set before applying the method')
        elif rewritten == None:
            raise TypeError('rewritten must be set before applying the method')
        if type(already_present) is not bool:
            raise TypeError('already_present must be a bool, not %s'
                            % type(already_present).__name__)
        xoriginal = original
        if original.startswith('tests=lib/'):
            xoriginal = substart('tests=lib/', 'lib/', original)
        lookedup, tmpflag = self.filesystem.lookup(xoriginal)
        tmpfile = self.tmpfilename(rewritten)
        sed_transform_build_aux_file = self.transformers.get('aux')
        sed_transform_main_lib_file = self.transformers.get('main')
        sed_transform_testsrelated_lib_file = self.transformers.get('tests')
        try:  # Try to copy lookedup file to tmpfile
            copyfile(lookedup, tmpfile)
            ensure_writable(tmpfile)
        except Exception as exc:
            raise GLError(15, lookedup) from exc
        # Don't process binary files with sed.
        if not (original.endswith('.class') or original.endswith('.mo')):
            transformer = None
            if original.startswith('lib/'):
                if sed_transform_main_lib_file:
                    transformer = sed_transform_main_lib_file
            elif original.startswith('build-aux/'):
                if sed_transform_build_aux_file:
                    transformer = sed_transform_build_aux_file
            elif original.startswith('tests=lib/'):
                if sed_transform_testsrelated_lib_file:
                    transformer = sed_transform_testsrelated_lib_file
            if transformer != None:
                # Read the file that we looked up.
                with open(lookedup, mode='r', newline='\n', encoding='utf-8') as file:
                    src_data = file.read()
                # Write the transformed data to the temporary file.
                with open(tmpfile, mode='w', newline='\n', encoding='utf-8') as file:
                    file.write(re.sub(transformer[0], transformer[1], src_data))
        path = joinpath(self.config['destdir'], rewritten)
        if os.path.isfile(path):
            # The file already exists.
            self.update(lookedup, tmpflag, tmpfile, already_present)
        else:  # if not os.path.isfile(path)
            # Install the file.
            # Don't protest if the file should be there but isn't: it happens
            # frequently that developers don't put autogenerated files under version control.
            self.add(lookedup, tmpflag, tmpfile)
            self.addFile(rewritten)
        if os.path.isfile(tmpfile):
            os.remove(tmpfile)

    def super_update(self, basename: str, tmpfile: str) -> tuple[str, str, int]:
        '''Move tmpfile to destdir/basename path, making a backup of it.
        Returns tuple, which contains basename, backupname and status.
          0: tmpfile is the same as destfile;
          1: tmpfile was used to update destfile;
          2: destfile was created, because it didn't exist.'''
        backupname = '%s~' % basename
        basepath = joinpath(self.config['destdir'], basename)
        backuppath = joinpath(self.config['destdir'], backupname)
        if os.path.isfile(basepath):
            if filecmp.cmp(basepath, tmpfile):
                result_flag = 0
            else:  # if not filecmp.cmp(basepath, tmpfile)
                result_flag = 1
                if not self.config['dryrun']:
                    if os.path.isfile(backuppath):
                        os.remove(backuppath)
                    movefile(basepath, backuppath)
                    movefile(tmpfile, basepath)
                else:  # if self.config['dryrun']
                    os.remove(tmpfile)
        else:  # if not os.path.isfile(basepath)
            result_flag = 2
            if not self.config['dryrun']:
                if os.path.isfile(basepath):
                    os.remove(basepath)
                movefile(tmpfile, basepath)
            else:  # if self.config['dryrun']
                os.remove(tmpfile)
        result = tuple([basename, backupname, result_flag])
        return result



# Copyright (C) 2002-2026 Free Software Foundation, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

#===============================================================================
# Define global imports
#===============================================================================
import os
import re
import sys
import shlex
import subprocess as sp
from pathlib import Path
from .constants import (
    DIRS,
    TESTS,
    UTILS,
    combine_lines,
    execute,
    ensure_writable,
    force_output,
    hardlink,
    joinpath,
    link_relative,
    lines_to_multiline,
    movefile,
    copyfile,
    substart,
    bold_escapes,
    relinverse,
    rmtree,
)
from .functions import rewrite_file_name
from .enums import CopyAction
from .GLError import GLError
from .GLConfig import GLConfig
from .GLModuleSystem import GLModuleTable, GLModuleSystem
from .GLFileSystem import GLFileSystem
from .GLMakefileTable import GLMakefileTable
from .GLEmiter import GLEmiter
from .GLFileTable import GLFileTable


def _patch_test_driver() -> None:
    '''Patch the test-driver script in testdirs.'''
    test_driver = joinpath('build-aux', 'test-driver')
    print('patching file %s' % test_driver)
    diffs = [ joinpath(DIRS['root'], name)
              for name in [joinpath('build-aux', 'test-driver.diff'),
                           joinpath('build-aux', 'test-driver-1.16.3.diff')] ]
    patched = False
    for diff in diffs:
        command = f'patch {shlex.quote(test_driver)} < {shlex.quote(diff)}'
        try:
            result = sp.call(command, shell=True, stdout=sp.DEVNULL, stderr=sp.DEVNULL)
        except OSError as exc:
            if os.path.isfile(f'{test_driver}.orig'):
                os.remove(f'{test_driver}.orig')
            if os.path.isfile(f'{test_driver}.rej'):
                os.remove(f'{test_driver}.rej')
            raise GLError(20, None) from exc
        if result == 0:
            patched = True
            break
        if os.path.isfile(f'{test_driver}.orig'):
            os.remove(f'{test_driver}.orig')
        if os.path.isfile(f'{test_driver}.rej'):
            os.remove(f'{test_driver}.rej')
    if not patched:
        raise GLError(20, None)
    if os.path.isfile(f'{test_driver}.orig'):
        os.remove(f'{test_driver}.orig')


#===============================================================================
# Define GLTestDir class
#===============================================================================
class GLTestDir:
    '''GLTestDir class is used to create a scratch package with the given
    list of the modules.'''

    config: GLConfig
    testdir: str
    emitter: GLEmiter
    filesystem: GLFileSystem
    modulesystem: GLModuleSystem
    makefiletable: GLMakefileTable

    def __init__(self, config: GLConfig, testdir: str) -> None:
        '''Create new GLTestDir instance.'''
        if type(config) is not GLConfig:
            raise TypeError('config must be a GLConfig, not %s'
                            % type(config).__name__)
        if type(testdir) is not str:
            raise TypeError('testdir must be a string, not %s'
                            % type(testdir).__name__)
        self.config = config
        self.testdir = os.path.normpath(testdir)
        # Don't overwrite the directory.
        if os.path.exists(self.testdir):
            raise GLError(22, self.testdir)
        # Try to create directory.
        try:
            os.mkdir(self.testdir)
        except Exception as exc:
            raise GLError(19, self.testdir) from exc
        self.emitter = GLEmiter(self.config)
        self.filesystem = GLFileSystem(self.config)
        self.modulesystem = GLModuleSystem(self.config)
        self.makefiletable = GLMakefileTable(self.config)

        # Subdirectory names.
        self.config.setSourceBase('gllib')
        self.config.setM4Base('glm4')
        self.config.setDocBase('gldoc')
        self.config.setTestsBase('gltests')
        self.config.setMacroPrefix('gl')
        self.config.resetPoBase()
        self.config.resetPoDomain()
        self.config.resetWitnessCMacro()
        self.config.resetVCFiles()

    def execute(self) -> None:
        '''Create a scratch package with the given modules.'''
        auxdir = self.config['auxdir']
        sourcebase = self.config['sourcebase']
        m4base = self.config['m4base']
        testsbase = self.config['testsbase']
        libname = self.config['libname']
        libtool = self.config['libtool']
        single_configure = self.config['single_configure']
        macro_prefix = self.config['macro_prefix']
        verbose = self.config['verbosity']

        specified_modules = self.config['modules']
        if len(specified_modules) == 0:
            # All modules together, except those that are not usable in a testdir.
            specified_modules = self.modulesystem.list()
            specified_modules = [ module_name
                                  for module_name in specified_modules
                                  if self.modulesystem.find(module_name) is not None \
                                      and self.modulesystem.find(module_name).getUsabilityInTestdir() != 'no' ]

        # Canonicalize the list of specified modules.
        modules = set()
        for module_name in specified_modules:
            module = self.modulesystem.find(module_name)
            if module is not None:
                if module.getUsabilityInTestdir() == 'no':
                    if self.config['errors']:
                        raise GLError(24, module_name)
                    else:  # if not self.config['errors']
                        sys.stderr.write('gnulib-tool: warning: ')
                        sys.stderr.write('module %s cannot be used in a testdir\n' % module_name)
                else:
                    modules.add(module)
        specified_modules = sorted(modules)

        # Test modules which invoke AC_CONFIG_FILES cannot be used with --with-tests
        # (without --two-configures). Avoid them.
        inctests = self.config.checkInclTestCategory(TESTS['tests'])
        if inctests and self.config.checkSingleConfigure():
            self.config.addAvoid('havelib-tests')

        # Now it's time to create the GLModuleTable.
        moduletable = GLModuleTable(self.config,
                                    True,
                                    self.config.checkInclTestCategory(TESTS['all-tests']))

        # When computing transitive closures, don't consider $module to depend on
        # $module-tests. Need this because tests are implicitly GPL and may depend
        # on GPL modules - therefore we don't want a warning in this case.
        saved_inctests = self.config.checkInclTestCategory(TESTS['tests'])
        self.config.disableInclTestCategory(TESTS['tests'])
        for requested_module in specified_modules:
            requested_licence = requested_module.getLicense()
            if requested_licence != 'GPL':
                # Here we use moduletable.transitive_closure([module]), not
                # just module.getDependencies, so that we also detect weird
                # situations like an LGPL module which depends on a GPLed build
                # tool module which depends on a GPL module.
                modules = moduletable.transitive_closure([requested_module])
                for module in modules:
                    license = module.getLicense()
                    if license not in ['GPLv2+ build tool', 'GPLed build tool',
                                       'public domain', 'unlimited', 'unmodifiable license text']:
                        incompatible = False
                        if requested_licence == 'GPLv3+' or requested_licence == 'GPL':
                            if license not in ['LGPLv2+', 'LGPLv3+ or GPLv2+', 'LGPLv3+', 'LGPL', 'GPLv2+', 'GPLv3+', 'GPL']:
                                incompatible = True
                        elif requested_licence == 'GPLv2+':
                            if license not in ['LGPLv2+', 'LGPLv3+ or GPLv2+', 'GPLv2+']:
                                incompatible = True
                        elif requested_licence == 'LGPLv3+' or requested_licence == 'LGPL':
                            if license not in ['LGPLv2+', 'LGPLv3+ or GPLv2+', 'LGPLv3+', 'LGPL']:
                                incompatible = True
                        elif requested_licence == 'LGPLv3+ or GPLv2+':
                            if license not in ['LGPLv2+', 'LGPLv3+ or GPLv2+']:
                                incompatible = True
                        elif requested_licence == 'LGPLv2+':
                            if license not in ['LGPLv2+']:
                                incompatible = True
                        if incompatible:
                            warningmsg = 'module %s depends on a module with an incompatible license: %s' % (requested_module, module)
                            sys.stderr.write('gnulib-tool: warning: %s\n' % warningmsg)
        self.config.setInclTestCategory(TESTS['tests'], saved_inctests)

        # Determine final module list.
        modules = moduletable.transitive_closure(specified_modules)
        final_modules = list(modules)

        # Show final module list.
        if verbose >= 0:
            (bold_on, bold_off) = bold_escapes()
            print('Module list with included dependencies (indented):')
            specified_modules_set = { module.name
                                      for module in specified_modules }
            for module in final_modules:
                if module.name in specified_modules_set:
                    print('  %s%s%s' % (bold_on, module.name, bold_off))
                else:  # if module.name not in specified_modules_set
                    print('    %s' % module.name)

        # Generate lists of the modules.
        if single_configure:
            # Determine main module list and tests-related module list separately.
            main_modules, tests_modules = \
                moduletable.transitive_closure_separately(specified_modules, final_modules)
            # Print main_modules and tests_modules.
            if verbose >= 1:
                print('Main module list:')
                for module in main_modules:
                    print('  %s' % module.name)
                print('Tests-related module list:')
                for module in tests_modules:
                    print('  %s' % module.name)
            # Determine whether a $testsbase/libtests.a is needed.
            libtests = False
            for module in tests_modules:
                files = module.getFiles()
                for file in files:
                    if file.startswith('lib/'):
                        libtests = True
                        break
            self.emitter.config.setLibtests(libtests)

        if single_configure:
            # Add the dummy module to the main module list if needed.
            main_modules = moduletable.add_dummy(main_modules)
            if libtests:  # if we need to use libtests.a
                # Add the dummy module to the tests-related module list if needed.
                tests_modules = moduletable.add_dummy(tests_modules)
        else:  # if not single_configure
            modules = moduletable.add_dummy(modules)

        # Show banner notice of every module.
        if single_configure:
            if verbose >= -1:
                for module in main_modules:
                    notice = module.getNotice().strip('\n')
                    if notice:
                        print('Notice from module %s:' % module.name)
                        pattern = re.compile(r'^(.*)$', re.M)
                        notice = pattern.sub(r'  \1', notice)
                        print(notice)
        else:  # if not single_configure
            if verbose >= -1:
                for module in modules:
                    notice = module.getNotice().strip('\n')
                    if notice:
                        print('Notice from module %s:' % module.name)
                        pattern = re.compile(r'^(.*)$', re.M)
                        notice = pattern.sub(r'  \1', notice)
                        print(notice)

        # Determine final file list.
        if single_configure:
            main_filelist, tests_filelist = \
                moduletable.filelist_separately(main_modules, tests_modules)
        else:  # if not single_configure
            main_modules = modules
            tests_modules = [ module
                              for module in modules
                              if module.repeatModuleInTests() ]
            main_filelist, tests_filelist = \
                moduletable.filelist_separately(main_modules, tests_modules)

        filelist = sorted(set(main_filelist + tests_filelist))

        # Print list of files.
        if verbose >= 0:
            print('File list:')
            for file in filelist:
                if file.startswith('tests=lib/'):
                    rest = file[10:]
                    print('  lib/%s -> tests/%s' % (rest, rest))
                else:
                    print('  %s' % file)

        # Add files for which the copy in gnulib is newer than the one that
        # "automake --add-missing --copy" would provide.
        filelist = sorted(set(filelist + ['build-aux/config.guess', 'build-aux/config.sub']))

        # new_table is a table with two columns: (rewritten-file-name original-file-name),
        # representing the files after this gnulib-tool invocation.
        new_table = { (rewrite_file_name(file_name, self.config, True), file_name)
                      for file_name in filelist }

        # Setup the file table.
        filetable = GLFileTable(filelist)
        filetable.new_files = sorted(new_table, key=lambda pair: pair[0])

        # Create directories.
        directories = sorted({ joinpath(self.testdir, os.path.dirname(pair[0]))
                               for pair in filetable.new_files })
        for directory in directories:
            if not os.path.isdir(directory):
                os.makedirs(directory)

        # Copy files or make symbolic links or hard links.
        for (dest, src) in filetable.new_files:
            destpath = joinpath(self.testdir, dest)
            if src.startswith('tests=lib/'):
                src = substart('tests=lib/', 'lib/', src)
            lookedup, flag = self.filesystem.lookup(src)
            if os.path.isfile(destpath):
                os.remove(destpath)
            if flag:
                copyfile(lookedup, destpath)
                ensure_writable(destpath)
            else:  # if not flag
                if self.filesystem.shouldLink(src, lookedup) == CopyAction.Symlink:
                    link_relative(lookedup, destpath)
                elif self.filesystem.shouldLink(src, lookedup) == CopyAction.Hardlink:
                    hardlink(lookedup, destpath)
                else:
                    copyfile(lookedup, destpath)
                    ensure_writable(destpath)

        # Create $sourcebase/Makefile.am.
        for_test = True
        directory = joinpath(self.testdir, sourcebase)
        if not os.path.isdir(directory):
            os.mkdir(directory)
        destfile = joinpath(directory, 'Makefile.am')
        if single_configure:
            emit = self.emitter.lib_Makefile_am(destfile, main_modules,
                                                moduletable, self.makefiletable, '', for_test)
        else:  # if not single_configure
            emit = self.emitter.lib_Makefile_am(destfile, modules,
                                                moduletable, self.makefiletable, '', for_test)
        with open(destfile, mode='w', newline='\n', encoding='utf-8') as file:
            file.write(emit)

        # Create $m4base/Makefile.am.
        directory = joinpath(self.testdir, m4base)
        if not os.path.isdir(directory):
            os.mkdir(directory)
        destfile = joinpath(directory, 'Makefile.am')
        emit = '## Process this file with automake to produce Makefile.in.\n\n'
        emit += 'EXTRA_DIST =\n'
        for file in filetable.all_files:
            if file.startswith('m4/'):
                file = substart('m4/', '', file)
                emit += 'EXTRA_DIST += %s\n' % file
        with open(destfile, mode='w', newline='\n', encoding='utf-8') as file:
            file.write(emit)

        subdirs = [sourcebase, m4base]
        subdirs_with_configure_ac = []

        inctests = self.config.checkInclTestCategory(TESTS['tests'])
        if inctests:
            directory = joinpath(self.testdir, testsbase)
            if not os.path.isdir(directory):
                os.mkdir(directory)
            if single_configure:
                # Create $testsbase/Makefile.am.
                destfile = joinpath(directory, 'Makefile.am')
                witness_macro = '%stests_WITNESS' % macro_prefix
                emit = self.emitter.tests_Makefile_am(destfile, tests_modules, moduletable,
                                                      self.makefiletable, witness_macro, for_test)
                with open(destfile, mode='w', newline='\n', encoding='utf-8') as file:
                    file.write(emit)
            else:  # if not single_configure
                # Create $testsbase/Makefile.am.
                destfile = joinpath(directory, 'Makefile.am')
                libtests = False
                self.config.setLibtests(False)
                emit = self.emitter.tests_Makefile_am(destfile, modules, moduletable,
                                                      self.makefiletable, '', for_test)
                with open(destfile, mode='w', newline='\n', encoding='utf-8') as file:
                    file.write(emit)
                # Viewed from the $testsbase subdirectory, $auxdir is different.
                emit = ''
                saved_auxdir = auxdir
                auxdir = os.path.normpath(joinpath(relinverse(testsbase), auxdir))
                self.config.setAuxDir(auxdir)
                # Create $testsbase/configure.ac.
                emit += '# Process this file with autoconf '
                emit += 'to produce a configure script.\n'
                emit += 'AC_INIT([dummy], [0])\n'
                emit += 'AC_CONFIG_AUX_DIR([%s])\n' % auxdir
                emit += 'AM_INIT_AUTOMAKE\n\n'
                emit += 'AC_CONFIG_HEADERS([config.h])\n\n'
                emit += 'AC_PROG_CC\n'
                emit += 'AC_PROG_INSTALL\n'
                emit += 'AC_PROG_MAKE_SET\n'
                emit += self.emitter.preEarlyMacros(False, '', modules)
                snippets = []
                for module in modules:
                    if module.name in ['gnumakefile', 'maintainer-makefile']:
                        # These are meant to be used only in the top-level directory.
                        pass
                    # if module.name not in ['gnumakefile', 'maintainer-makefile']
                    else:
                        snippet = module.getAutoconfEarlySnippet()
                        lines = [ line
                                  for line in snippet.split('\n')
                                  if line.strip() ]
                        snippet = lines_to_multiline(lines)
                        pattern = re.compile(r'AC_REQUIRE\(\[([^()]*)]\)', re.M)
                        snippet = pattern.sub(r'\1', snippet)
                        snippet = snippet.strip()
                        snippets.append(snippet)
                snippets = [ snippet
                             for snippet in snippets
                             if snippet.strip()]
                emit += lines_to_multiline(snippets)
                if libtool:
                    emit += 'LT_INIT([win32-dll])\n'
                    emit += 'LT_LANG([C++])\n'
                    emit += 'AM_CONDITIONAL([GL_COND_LIBTOOL], [true])\n'
                    emit += 'gl_cond_libtool=true\n'
                else:  # if not libtool
                    emit += 'AM_CONDITIONAL([GL_COND_LIBTOOL], [false])\n'
                    emit += 'gl_cond_libtool=false\n'
                    emit += 'gl_libdeps=\n'
                    emit += 'gl_ltlibdeps=\n'
                # Wrap the set of autoconf snippets into an autoconf macro that is then
                # invoked. This is needed because autoconf does not support AC_REQUIRE
                # at the top level:
                #   error: AC_REQUIRE(gt_CSHARPCOMP): cannot be used outside of an
                #     AC_DEFUN'd macro
                # but we want the AC_REQUIRE to have its normal meaning (provide one
                # expansion of the required macro before the current point, and only
                # one expansion total).
                emit += 'AC_DEFUN([gl_INIT], [\n'
                replace_auxdir = True
                emit += "gl_m4_base='../%s'\n" % m4base
                emit += self.emitter.initmacro_start(macro_prefix, True)
                # We don't have explicit ordering constraints between the various
                # autoconf snippets. It's cleanest to put those of the library before
                # those of the tests.
                emit += self.emitter.shellvars_init(True, f'../{sourcebase}')
                emit += self.emitter.autoconfSnippets(modules, modules, moduletable,
                                                      lambda module: module.isNonTests(),
                                                      False, False, False, replace_auxdir)
                emit += self.emitter.shellvars_init(True, '.')
                emit += self.emitter.autoconfSnippets(modules, modules, moduletable,
                                                      lambda module: module.isTests(),
                                                      False, False, False, replace_auxdir)
                emit += self.emitter.initmacro_end(macro_prefix, True)
                # _LIBDEPS and _LTLIBDEPS variables are not needed if this library is
                # created using libtool, because libtool already handles the
                # dependencies.
                if not libtool:
                    libname_upper = libname.upper().replace('-', '_')
                    emit += '  %s_LIBDEPS="$gl_libdeps"\n' % libname_upper
                    emit += '  AC_SUBST([%s_LIBDEPS])\n' % libname_upper
                    emit += '  %s_LTLIBDEPS="$gl_ltlibdeps"\n' % libname_upper
                    emit += '  AC_SUBST([%s_LTLIBDEPS])\n' % libname_upper
                emit += '])\n'
                # FIXME use $sourcebase or $testsbase?
                emit += self.emitter.initmacro_done(macro_prefix, sourcebase)
                emit += '\ngl_INIT\n\n'
                # Usually $testsbase/config.h will be a superset of config.h. Verify
                # this by "merging" config.h into $testsbase/config.h; look out for gcc
                # warnings.
                emit += 'AH_TOP([#include \"../config.h\"])\n\n'
                emit += 'AC_CONFIG_FILES([Makefile])\n'
                emit += 'AC_OUTPUT\n'
                path = joinpath(self.testdir, testsbase, 'configure.ac')
                with open(path, mode='w', newline='\n', encoding='utf-8') as file:
                    file.write(emit)

                # Restore changed variables.
                self.config.setAuxDir(saved_auxdir)
                auxdir = self.config['auxdir']
                subdirs_with_configure_ac.append(testsbase)

            subdirs.append(testsbase)

        # Create Makefile.am.
        emit = '## Process this file with automake to produce Makefile.in.\n\n'
        emit += 'AUTOMAKE_OPTIONS = 1.14 foreign\n\n'
        emit += 'SUBDIRS = %s\n\n' % ' '.join(subdirs)
        emit += 'ACLOCAL_AMFLAGS = -I %s\n' % m4base
        path = joinpath(self.testdir, 'Makefile.am')
        with open(path, mode='w', newline='\n', encoding='utf-8') as file:
            file.write(emit)

        # Create configure.ac
        emit = '# Process this file with autoconf '
        emit += 'to produce a configure script.\n'
        emit += 'AC_INIT([dummy], [0])\n'
        if auxdir != '.':
            emit += 'AC_CONFIG_AUX_DIR([%s])\n' % auxdir
        emit += 'AM_INIT_AUTOMAKE\n\n'
        emit += 'AC_CONFIG_HEADERS([config.h])\n\n'
        emit += 'AC_PROG_CC\n'
        emit += 'AC_PROG_INSTALL\n'
        emit += 'AC_PROG_MAKE_SET\n\n'
        emit += '# For autobuild.\n'
        emit += 'AC_CANONICAL_BUILD\n'
        emit += 'AC_CANONICAL_HOST\n\n'
        emit += 'm4_pattern_forbid([^gl_[A-Z]])dnl the gnulib macro namespace\n'
        emit += 'm4_pattern_allow([^gl_ES$])dnl a valid locale name\n'
        emit += 'm4_pattern_allow([^gl_LIBOBJS$])dnl a variable\n'
        emit += 'm4_pattern_allow([^gl_LTLIBOBJS$])dnl a variable\n'
        emit += self.emitter.preEarlyMacros(False, '', modules)
        snippets = []
        for module in final_modules:
            if single_configure:
                solution = True
            else:  # if not single_configure
                solution = module.isNonTests()
            if solution:
                snippet = module.getAutoconfEarlySnippet()
                lines = [ line
                          for line in snippet.split('\n')
                          if line.strip() ]
                snippet = lines_to_multiline(lines)
                pattern = re.compile(r'AC_REQUIRE\(\[([^()]*)]\)', re.M)
                snippet = pattern.sub(r'\1', snippet)
                snippet = snippet.strip()
                snippets.append(snippet)
        snippets = [ snippet
                     for snippet in snippets
                     if snippet.strip() ]
        emit += lines_to_multiline(snippets)
        if libtool:
            emit += 'LT_INIT([win32-dll])\n'
            emit += 'LT_LANG([C++])\n'
            emit += 'AM_CONDITIONAL([GL_COND_LIBTOOL], [true])\n'
            emit += 'gl_cond_libtool=true\n'
        else:  # if not libtool
            emit += 'AM_CONDITIONAL([GL_COND_LIBTOOL], [false])\n'
            emit += 'gl_cond_libtool=false\n'
            emit += 'gl_libdeps=\n'
            emit += 'gl_ltlibdeps=\n'
        # Wrap the set of autoconf snippets into an autoconf macro that is then
        # invoked. This is needed because autoconf does not support AC_REQUIRE
        # at the top level:
        #   error: AC_REQUIRE(gt_CSHARPCOMP): cannot be used outside of an
        #     AC_DEFUN'd macro
        # but we want the AC_REQUIRE to have its normal meaning (provide one
        # expansion of the required macro before the current point, and only one
        # expansion total).
        emit += 'AC_DEFUN([gl_INIT], [\n'
        if auxdir != 'build-aux':
            replace_auxdir = True
        else:  # auxdir == 'build-aux'
            replace_auxdir = False
        emit += 'gl_m4_base=\'%s\'\n' % m4base
        emit += self.emitter.initmacro_start(macro_prefix, False)
        emit += self.emitter.shellvars_init(False, sourcebase)
        if single_configure:
            emit += self.emitter.autoconfSnippets(main_modules, main_modules, moduletable,
                                                  lambda module: True,
                                                  True, False, False, replace_auxdir)
        else:  # if not single_configure
            emit += self.emitter.autoconfSnippets(modules, modules, moduletable,
                                                  lambda module: module.isNonTests(),
                                                  True, False, False, replace_auxdir)
        emit += self.emitter.initmacro_end(macro_prefix, False)
        if single_configure:
            emit += '  gltests_libdeps=\n'
            emit += '  gltests_ltlibdeps=\n'
            emit += self.emitter.initmacro_start('%stests' % macro_prefix, True)
            emit += self.emitter.shellvars_init(True, testsbase)
            # Define a tests witness macro.
            emit += '  %stests_WITNESS=IN_GNULIB_TESTS\n' % macro_prefix
            emit += '  AC_SUBST([%stests_WITNESS])\n' % macro_prefix
            emit += '  gl_module_indicator_condition=$%stests_WITNESS\n' % macro_prefix
            emit += '  m4_pushdef([gl_MODULE_INDICATOR_CONDITION], '
            emit += '[$gl_module_indicator_condition])\n'
            snippets = self.emitter.autoconfSnippets(tests_modules, main_modules + tests_modules,
                                                     moduletable, lambda module: True,
                                                     True, False, False, replace_auxdir)
            emit += snippets
            emit += '  m4_popdef([gl_MODULE_INDICATOR_CONDITION])\n'
            emit += self.emitter.initmacro_end('%stests' % macro_prefix, True)
        # _LIBDEPS and _LTLIBDEPS variables are not needed if this library is
        # created using libtool, because libtool already handles the dependencies.
        if not libtool:
            libname_upper = libname.upper().replace('-', '_')
            emit += '  %s_LIBDEPS="$gl_libdeps"\n' % libname_upper
            emit += '  AC_SUBST([%s_LIBDEPS])\n' % libname_upper
            emit += '  %s_LTLIBDEPS="$gl_ltlibdeps"\n' % libname_upper
            emit += '  AC_SUBST([%s_LTLIBDEPS])\n' % libname_upper
        if single_configure and libtests:
            emit += '  LIBTESTS_LIBDEPS="$gltests_libdeps"\n'
            emit += '  AC_SUBST([LIBTESTS_LIBDEPS])\n'
        emit += '])\n'
        emit += self.emitter.initmacro_done(macro_prefix, sourcebase)
        if single_configure:
            emit += self.emitter.initmacro_done('%stests' % macro_prefix, testsbase)
        emit += '\ngl_INIT\n\n'
        if subdirs_with_configure_ac:
            emit += 'AC_CONFIG_SUBDIRS([%s])\n' % ' '.join(subdirs_with_configure_ac)
        makefiles = ['Makefile']
        for directory in subdirs:
            # For subdirs that have a configure.ac by their own, it's the subdir's
            # configure.ac which creates the subdir's Makefile.am, not this one.
            if not directory in subdirs_with_configure_ac:
                makefiles.append(joinpath(directory, 'Makefile'))
        emit += 'AC_CONFIG_FILES([%s])\n' % ' '.join(makefiles)
        emit += 'AC_OUTPUT\n'
        path = joinpath(self.testdir, 'configure.ac')
        with open(path, mode='w', newline='\n', encoding='utf-8') as file:
            file.write(emit)

        # Create autogenerated files.
        # Do not use "${AUTORECONF} --force --install", because it may invoke
        # autopoint, which brings in older versions of some of our .m4 files.
        force_output()
        os.chdir(self.testdir)
        # gettext
        if os.path.isfile(joinpath(m4base, 'gettext.m4')):
            args = [UTILS['autopoint'], '--force']
            execute(args, verbose)
            for src in os.listdir(m4base):
                src = joinpath(m4base, src)
                if src.endswith('.m4~'):
                    dest = src[:-1]
                    if os.path.isfile(dest):
                        os.remove(dest)
                    movefile(src, dest)
        # libtoolize
        if libtool:
            args = [UTILS['libtoolize'], '--copy']
            execute(args, verbose)
        # aclocal
        args = [UTILS['aclocal'], '-I', m4base]
        execute(args, verbose)
        if not os.path.isdir('build-aux'):
            print('executing mkdir build-aux')
            os.mkdir('build-aux')
        # autoconf
        args = [UTILS['autoconf']]
        execute(args, verbose)
        # autoheader
        args = [UTILS['autoheader']]
        execute(args, verbose)
        # Explicit 'touch config.h.in': see <https://savannah.gnu.org/support/index.php?109406>.
        print('executing touch config.h.in')
        Path('config.h.in').touch()
        # automake
        args = [UTILS['automake'], '--add-missing', '--copy']
        execute(args, verbose)
        rmtree('autom4te.cache')
        os.chdir(DIRS['cwd'])
        if inctests and not single_configure:
            # Do not use "${AUTORECONF} --force --install", because it may invoke
            # autopoint, which brings in older versions of some of our .m4 files.
            os.chdir(joinpath(self.testdir, testsbase))
            # gettext
            if os.path.isfile(joinpath(m4base, 'gettext.m4')):
                args = [UTILS['autopoint'], '--force']
                execute(args, verbose)
                for src in os.listdir(m4base):
                    src = joinpath(m4base, src)
                    if src.endswith('.m4~'):
                        dest = src[:-1]
                        if os.path.isfile(dest):
                            os.remove(dest)
                        movefile(src, dest)
            # aclocal
            args = [UTILS['aclocal'], '-I', joinpath('..', m4base)]
            execute(args, verbose)
            if not os.path.isdir(joinpath('../build-aux')):
                print('executing mkdir ../build-aux')
                os.mkdir('../build-aux')
            # autoconf
            args = [UTILS['autoconf']]
            execute(args, verbose)
            # autoheader
            args = [UTILS['autoheader']]
            execute(args, verbose)
            # Explicit 'touch config.h.in': see <https://savannah.gnu.org/support/index.php?109406>.
            print('executing touch config.h.in')
            Path('config.h.in').touch()
            # automake
            args = [UTILS['automake'], '--add-missing', '--copy']
            execute(args, verbose)
            rmtree('autom4te.cache')
            os.chdir(DIRS['cwd'])

        # Need to run configure and make once, to create built files that are to be
        # distributed (such as parse-datetime.c).
        path = joinpath(self.testdir, sourcebase, 'Makefile.am')
        with open(path, mode='r', newline='\n', encoding='utf-8') as file:
            snippet = file.read()
        snippet = combine_lines(snippet)

        # Extract the value of "CLEANFILES += ..." and "MOSTLYCLEANFILES += ...".
        regex_find = []
        pattern = re.compile(r'^CLEANFILES[\t ]*\+=(.*)$', re.M)
        regex_find += pattern.findall(snippet)
        pattern = re.compile(r'^MOSTLYCLEANFILES[\t ]*\+=(.*)$', re.M)
        regex_find += pattern.findall(snippet)
        regex_find = [ line.strip()
                       for line in regex_find
                       if line.strip() ]
        cleaned_files = []
        for part in regex_find:
            cleaned_files += \
                [ line.strip()
                  for line in part.split(' ')
                  if line.strip() ]

        # Extract the value of "BUILT_SOURCES += ...". Remove variable references
        # such $(FOO_H) because they don't refer to distributed files.
        regex_find = []
        pattern = re.compile(r'^BUILT_SOURCES[\t ]*\+=(.*)$', re.M)
        regex_find += pattern.findall(snippet)
        regex_find = [ line.strip()
                       for line in regex_find
                       if line.strip()]
        built_sources = []
        for part in regex_find:
            built_sources += \
                [ line.strip()
                  for line in part.split(' ')
                  if line.strip()]
        built_sources = [ line
                          for line in built_sources
                          if not bool(re.compile(r'[$]\([A-Za-z0-9_]*\)$').findall(line)) ]
        distributed_built_sources = [ file
                                      for file in built_sources
                                      if file not in cleaned_files ]

        tests_distributed_built_sources = []
        if inctests:
            # Likewise for built files in the $testsbase directory.
            path = joinpath(self.testdir, testsbase, 'Makefile.am')
            with open(path, mode='r', newline='\n', encoding='utf-8') as file:
                snippet = file.read()
            snippet = combine_lines(snippet)

            # Extract the value of "CLEANFILES += ..." and "MOSTLYCLEANFILES += ...".
            regex_find = []
            pattern = re.compile(r'^CLEANFILES[\t ]*\+=(.*)$', re.M)
            regex_find += pattern.findall(snippet)
            pattern = re.compile(r'^MOSTLYCLEANFILES[\t ]*\+=(.*)$', re.M)
            regex_find += pattern.findall(snippet)
            regex_find = [ line.strip()
                           for line in regex_find
                           if line.strip() ]
            tests_cleaned_files = []
            for part in regex_find:
                tests_cleaned_files += \
                    [ line.strip()
                      for line in part.split(' ')
                      if line.strip() ]

            # Extract the value of "BUILT_SOURCES += ...". Remove variable references
            # such $(FOO_H) because they don't refer to distributed files.
            regex_find = []
            pattern = re.compile(r'^BUILT_SOURCES[\t ]*\+=(.*)$', re.M)
            regex_find += pattern.findall(snippet)
            regex_find = [ line.strip()
                           for line in regex_find
                           if line.strip() ]
            tests_built_sources = []
            for part in regex_find:
                tests_built_sources += \
                    [ line.strip()
                      for line in part.split(' ')
                      if line.strip() ]
            tests_built_sources = [ line
                                    for line in tests_built_sources
                                    if not bool(re.compile(r'[$]\([A-Za-z0-9_]*\)$').findall(line)) ]
            tests_distributed_built_sources = [ file
                                                for file in tests_built_sources
                                                if file not in tests_cleaned_files]

        os.chdir(self.testdir)
        if distributed_built_sources or tests_distributed_built_sources:
            force_output()
            sp.call('./configure')
            if distributed_built_sources:
                os.chdir(sourcebase)
                with open('Makefile', mode='a', newline='\n', encoding='utf-8') as file:
                    file.write('built_sources: $(BUILT_SOURCES)\n')
                args = [UTILS['make'],
                        'AUTOCONF=%s' % UTILS['autoconf'],
                        'AUTOHEADER=%s' % UTILS['autoheader'],
                        'ACLOCAL=%s' % UTILS['aclocal'],
                        'AUTOMAKE=%s' % UTILS['automake'],
                        'AUTORECONF=%s' % UTILS['autoreconf'],
                        'built_sources']
                sp.call(args)
                os.chdir('..')
            if tests_distributed_built_sources:
                os.chdir(testsbase)
                with open('Makefile', mode='a', newline='\n', encoding='utf-8') as file:
                    file.write('built_sources: $(BUILT_SOURCES)\n')
                args = [UTILS['make'],
                        'AUTOCONF=%s' % UTILS['autoconf'],
                        'AUTOHEADER=%s' % UTILS['autoheader'],
                        'ACLOCAL=%s' % UTILS['aclocal'],
                        'AUTOMAKE=%s' % UTILS['automake'],
                        'AUTORECONF=%s' % UTILS['autoreconf'],
                        'built_sources']
                sp.call(args)
                os.chdir('..')
            args = [UTILS['make'],
                    'AUTOCONF=%s' % UTILS['autoconf'],
                    'AUTOHEADER=%s' % UTILS['autoheader'],
                    'ACLOCAL=%s' % UTILS['aclocal'],
                    'AUTOMAKE=%s' % UTILS['automake'],
                    'AUTORECONF=%s' % UTILS['autoreconf'],
                    'AUTOPOINT=%s' % UTILS['autopoint'],
                    'LIBTOOLIZE=%s' % UTILS['libtoolize'],
                    'distclean']
            sp.call(args)
        if os.path.isfile(joinpath('build-aux', 'test-driver')):
            _patch_test_driver()
        os.chdir(DIRS['cwd'])


#===============================================================================
# Define GLMegaTestDir class
#===============================================================================
class GLMegaTestDir:
    '''GLMegaTestDir class is used to create a mega scratch package with the
    given modules one by one and all together.'''

    config: GLConfig
    megatestdir: str
    modulesystem: GLModuleSystem

    def __init__(self, config: GLConfig, megatestdir: str) -> None:
        '''Create new GLTestDir instance.'''
        if type(config) is not GLConfig:
            raise TypeError('config must be a GLConfig, not %s'
                            % type(config).__name__)
        if type(megatestdir) is not str:
            raise TypeError('megatestdir must be a string, not %s'
                            % type(megatestdir).__name__)
        self.config = config
        self.megatestdir = os.path.normpath(megatestdir)
        # Don't overwrite the directory.
        if os.path.exists(self.megatestdir):
            raise GLError(22, self.megatestdir)
        # Try to create directory.
        try:
            os.mkdir(self.megatestdir)
        except Exception as exc:
            raise GLError(19, self.megatestdir) from exc
        self.modulesystem = GLModuleSystem(self.config)

    def execute(self) -> None:
        '''Create a mega scratch package with the given modules one by one
        and all together.'''
        auxdir = self.config['auxdir']
        verbose = self.config['verbosity']

        megasubdirs = []
        modules = [ self.modulesystem.find(m)
                    for m in self.config['modules'] ]
        if not modules:
            modules = self.modulesystem.list()
            modules = [ self.modulesystem.find(m)
                        for m in modules ]
        # Preserve ordering from the command-line, but remove duplicates.
        # This allows control over the SUBDIRS variable in the top-level Makefile.am.
        module_set = set(modules)
        modules = [ module
                    for module in modules
                    if module in module_set ]

        # First, all modules one by one.
        for module in modules:
            self.config.setModules([module.name])
            GLTestDir(self.config, joinpath(self.megatestdir, module.name)).execute()
            megasubdirs.append(module.name)

        # Then, all modules all together.
        # Except config-h, which breaks all modules which use HAVE_CONFIG_H.
        modules = [ module
                    for module in modules
                    if module.name != 'config-h' ]
        self.config.setModules([ module.name
                                 for module in modules ])
        GLTestDir(self.config, joinpath(self.megatestdir, 'ALL')).execute()
        megasubdirs.append('ALL')

        # Create autobuild.
        emit = ''
        repdict = dict()
        repdict['Jan'] = repdict['January'] = '01'
        repdict['Feb'] = repdict['February'] = '02'
        repdict['Mar'] = repdict['March'] = '03'
        repdict['Apr'] = repdict['April'] = '04'
        repdict['May'] = repdict['May'] = '05'
        repdict['Jun'] = repdict['June'] = '06'
        repdict['Jul'] = repdict['July'] = '07'
        repdict['Aug'] = repdict['August'] = '08'
        repdict['Sep'] = repdict['September'] = '09'
        repdict['Oct'] = repdict['October'] = '10'
        repdict['Nov'] = repdict['November'] = '11'
        repdict['Dec'] = repdict['December'] = '12'
        vc_witness = joinpath(DIRS['root'], '.git', 'refs', 'heads', 'master')
        if not os.path.isfile(vc_witness):
            vc_witness = joinpath(DIRS['root'], 'ChangeLog')
        mdate_sh = joinpath(DIRS['root'], 'build-aux', 'mdate-sh')
        args = ['sh', mdate_sh, vc_witness]
        cvsdate = sp.check_output(args).decode('UTF-8').strip()
        for key in repdict:
            if len(key) > 3:
                cvsdate = cvsdate.replace(key, repdict[key])
        for key in repdict:
            cvsdate = cvsdate.replace(key, repdict[key])
        cvsdate = ''.join([ date
                            for date in cvsdate.split(' ')
                            if date.strip() ])
        cvsdate = '%s%s%s' % (cvsdate[4:], cvsdate[2:4], cvsdate[:2])
        emit += '#!/bin/sh\n'
        emit += 'CVSDATE=%s\n' % cvsdate
        emit += ': ${MAKE=make}\n'
        emit += 'test -d logs || mkdir logs\n'
        emit += 'for module in %s; do\n' % ' '.join(megasubdirs)
        emit += '  echo "Working on module $module..."\n'
        emit += '  safemodule=`echo $module | sed -e \'s|/|-|g\'`\n'
        emit += '  (echo "To: gnulib@autobuild.josefsson.org"\n'
        emit += '   echo\n'
        emit += '   set -x\n'
        emit += '   : autobuild project... $module\n'
        emit += '   : autobuild revision... cvs-$CVSDATE-000000\n'
        emit += '   : autobuild timestamp... `date "+%Y%m%d-%H%M%S"`\n'
        emit += '   : autobuild hostname... `hostname`\n'
        emit += '   cd $module && ./configure $CONFIGURE_OPTIONS && $MAKE'
        emit += ' && $MAKE check && $MAKE distclean\n'
        emit += '   echo rc=$?\n'
        emit += '  ) 2>&1 | { if test -n "$AUTOBUILD_SUBST"; then '
        emit += 'sed -e "$AUTOBUILD_SUBST"; else cat; fi; } > logs/$safemodule\n'
        emit += 'done\n'
        path = joinpath(self.megatestdir, 'do-autobuild')
        with open(path, mode='w', newline='\n', encoding='utf-8') as file:
            file.write(emit)

        # Create Makefile.am.
        emit = '## Process this file with automake to produce Makefile.in.\n\n'
        emit += 'AUTOMAKE_OPTIONS = 1.14 foreign\n\n'
        emit += 'SUBDIRS = %s\n\n' % ' '.join(megasubdirs)
        emit += 'EXTRA_DIST = do-autobuild\n'
        path = joinpath(self.megatestdir, 'Makefile.am')
        with open(path, mode='w', newline='\n', encoding='utf-8') as file:
            file.write(emit)

        emit = '# Process this file with autoconf '
        emit += 'to produce a configure script.\n'
        emit += 'AC_INIT([dummy], [0])\n'
        if auxdir != '.':
            emit += 'AC_CONFIG_AUX_DIR([%s])\n' % auxdir
        emit += 'AM_INIT_AUTOMAKE\n\n'
        emit += 'AC_PROG_MAKE_SET\n\n'
        emit += 'AC_CONFIG_SUBDIRS([%s])\n' % ' '.join(megasubdirs)
        emit += 'AC_CONFIG_FILES([Makefile])\n'
        emit += 'AC_OUTPUT\n'
        path = joinpath(self.megatestdir, 'configure.ac')
        with open(path, mode='w', newline='\n', encoding='utf-8') as file:
            file.write(emit)

        # Create autogenerated files.
        force_output()
        os.chdir(self.megatestdir)
        args = [UTILS['aclocal']]
        execute(args, verbose)
        try:  # Try to make a directory
            if not os.path.isdir('build-aux'):
                print('executing mkdir build-aux')
                os.mkdir('build-aux')
        except Exception:
            pass
        args = [UTILS['autoconf']]
        execute(args, verbose)
        args = [UTILS['automake'], '--add-missing', '--copy']
        execute(args, verbose)
        rmtree('autom4te.cache')
        if os.path.isfile(joinpath('build-aux', 'test-driver')):
            _patch_test_driver()
        os.chdir(DIRS['cwd'])



#!/usr/bin/python3
# Main VCSToChangeLog script.
# Copyright (C) 2019-2026 Free Software Foundation, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

''' Generate a ChangeLog style output based on a VCS log.

This script takes two revisions as input and generates a ChangeLog style output
for all revisions between the two revisions.

This script is intended to be executed from the project parent directory.

The vcstocl directory has a file vcstocl_quirks.py that defines a
function called get_project_quirks that returns a object of class type
ProjectQuirks or a subclass of the same.  The definition of the ProjectQuirks
class is below and it specifies the properties that the project must set to
ensure correct parsing of its contents.

Among other things, ProjectQuirks specifies the VCS to read from; the default is
assumed to be git.  The script then studies the VCS log and for each change,
list out the nature of changes in the constituent files.

Each file type may have parser frontends that can read files and construct
objects that may be compared to determine the minimal changes that occurred in
each revision.  For files that do not have parsers, we may only know the nature
of changes at the top level depending on the information that the VCS stores.

The parser frontend must have a compare() method that takes the old and new
files as arrays of strings and prints the output in ChangeLog format.

Currently implemented VCS:

    git

Currently implemented frontends:

    C
'''
import sys
import os
import re
import argparse
from vcstocl.misc_util import *
from vcstocl import frontend_c
from vcstocl.vcs_git import *
from vcstocl.projectquirks import ProjectQuirks

debug = DebugUtil(False)

sys.path.append('/'.join([os.path.dirname(os.path.realpath(__file__)),
                'vcstocl']))

def main(repo, frontends, refs):
    ''' ChangeLog Generator Entry Point.
    '''
    commits = repo.list_commits(args.refs)
    for commit in commits:
        repo.list_changes(commit, frontends)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()

    parser.add_argument('refs', metavar='ref', type=str, nargs=2,
                     help='Refs to print ChangeLog entries between')

    parser.add_argument('-d', '--debug', required=False, action='store_true',
                     help='Run the file parser debugger.')

    parser.add_argument('-q', '--quirks', required=False, type=str,
                     help='Load a quirks file.')

    args = parser.parse_args()

    debug.debug = args.debug

    if len(args.refs) < 2:
        debug.eprint('Two refs needed to get a ChangeLog.')
        sys.exit(os.EX_USAGE)

    # Load quirks file.  We assume that the script is run from the top level source
    # directory.
    if args.quirks:
        import importlib.util
        spec = importlib.util.spec_from_file_location("vcstocl_quirks", args.quirks)
        vcstocl_quirks = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(vcstocl_quirks)
        project_quirks = vcstocl_quirks.get_project_quirks(debug)
    else:
        try:
            from vcstocl_quirks import *
            project_quirks = get_project_quirks(debug)
        except:
            project_quirks = ProjectQuirks()

    REPO = {'git': GitRepo(project_quirks.IGNORE_LIST, debug)}

    fe_c = frontend_c.Frontend(project_quirks, debug)
    FRONTENDS = {'.c': fe_c,
                 '.h': fe_c}

    main(REPO[project_quirks.repo], FRONTENDS, args.refs)



# General Utility functions.
# Copyright (C) 2019-2026 Free Software Foundation, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import sys
import os

class DebugUtil:
    debug = False
    def __init__(self, debug):
        self.debug = debug

    def eprint(self, *args, **kwargs):
        ''' Print to stderr.
        '''
        print(*args, file=sys.stderr, **kwargs)


    def print(self, *args, **kwargs):
        ''' Convenience function to print diagnostic information in the program.
        '''
        if self.debug:
            self.eprint(*args, **kwargs)


def decode(string):
    ''' Attempt to decode a string.

    Decode a string read from the source file.  The multiple attempts are needed
    due to the presence of the page break characters and some tests in locales.
    '''
    codecs = ['utf8', 'cp1252']

    for i in codecs:
        try:
            return string.decode(i)
        except UnicodeDecodeError:
            pass

    DebugUtil.eprint('Failed to decode: %s' % string)


def analyze_diff(filename, oldfile, newfile, frontends):
    ''' Parse the output of the old and new files and print the difference.

    For input files OLDFILE and NEWFILE with name FILENAME, generate reduced
    trees for them and compare them.  We limit our comparison to only C source
    files.
    '''
    name, ext = os.path.splitext(filename)

    if not ext in frontends.keys():
        return None
    else:
        frontend = frontends[ext]
        frontend.compare(oldfile, newfile)



# Copyright (C) 2002-2026 Free Software Foundation, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

'''An easy access to pygnulib constants.'''

from __future__ import unicode_literals
from __future__ import annotations

#===============================================================================
# Define global imports
#===============================================================================
import re
import os
import sys
import stat
import shutil
import tempfile
import subprocess as sp
import __main__ as interpreter

#===============================================================================
# Define module information
#===============================================================================
__all__ = []


#===============================================================================
# Define global constants
#===============================================================================
# Declare necessary variables
APP = dict()  # Application
DIRS = dict()  # Directories
UTILS = dict()  # Utilities
ENCS = dict()  # Encodings
MODES = dict()  # Modes
TESTS = dict()  # Tests

# Set ENCS dictionary
if not hasattr(interpreter, '__file__'):
    if sys.stdout.encoding != None:
        ENCS['default'] = sys.stdout.encoding
    else:  # sys.stdout.encoding == None
        ENCS['default'] = 'UTF-8'
else:  # if hasattr(interpreter, '__file__'):
    ENCS['default'] = 'UTF-8'
ENCS['system'] = sys.getfilesystemencoding()
ENCS['shell'] = sys.stdout.encoding
if ENCS['shell'] == None:
    ENCS['shell'] = 'UTF-8'

# Set APP dictionary
APP['path'] = os.path.realpath(sys.argv[0])                 # file name of <gnulib>/.gnulib-tool.py
APP['root'] = os.path.dirname(APP['path'])                  # file name of <gnulib>
APP['name'] = os.path.join(APP['root'], 'gnulib-tool.py')

# Set DIRS directory
DIRS['cwd'] = os.getcwd()
def init_DIRS(gnulib_dir: str) -> None:
    DIRS['root'] = gnulib_dir
    DIRS['build-aux'] = os.path.join(gnulib_dir, 'build-aux')
    DIRS['config'] = os.path.join(gnulib_dir, 'config')
    DIRS['doc'] = os.path.join(gnulib_dir, 'doc')
    DIRS['lib'] = os.path.join(gnulib_dir, 'lib')
    DIRS['m4'] = os.path.join(gnulib_dir, 'm4')
    DIRS['modules'] = os.path.join(gnulib_dir, 'modules')
    DIRS['tests'] = os.path.join(gnulib_dir, 'tests')
    DIRS['git'] = os.path.join(gnulib_dir, '.git')
    DIRS['cvs'] = os.path.join(gnulib_dir, 'CVS')

# Set MODES dictionary
MODES = \
    {
        'import': 0,
        'add-import': 1,
        'remove-import': 2,
        'update': 3,
    }
MODES['verbose-min'] = -2
MODES['verbose-default'] = 0
MODES['verbose-max'] = 2

# Define TESTS categories
TESTS = \
    {
        'tests':             0,
        'c++-test':          1,
        'cxx-test':          1,
        'c++-tests':         1,
        'cxx-tests':         1,
        'longrunning-test':  2,
        'longrunning-tests': 2,
        'privileged-test':   3,
        'privileged-tests':  3,
        'unportable-test':   4,
        'unportable-tests':  4,
        'all-test':          5,
        'all-tests':         5,
    }

# Define AUTOCONF minimum version
DEFAULT_AUTOCONF_MINVERSION = 2.64
# You can set AUTOCONFPATH to empty if autoconf ≥ 2.64 is already in your PATH
AUTOCONFPATH = ''
# You can set AUTOMAKEPATH to empty if automake ≥ 1.14 is already in your PATH
AUTOMAKEPATH = ''
# You can set GETTEXTPATH to empty if autopoint ≥ 0.15 is already in your PATH
GETTEXTPATH = ''
# You can set LIBTOOLPATH to empty if libtoolize 2.x is already in your PATH
LIBTOOLPATH = ''

# You can also set the variable AUTOCONF individually
if AUTOCONFPATH:
    UTILS['autoconf'] = AUTOCONFPATH + 'autoconf'
else:
    if os.getenv('AUTOCONF'):
        UTILS['autoconf'] = os.getenv('AUTOCONF')
    else:
        UTILS['autoconf'] = 'autoconf'

# You can also set the variable AUTORECONF individually
if AUTOCONFPATH:
    UTILS['autoreconf'] = AUTOCONFPATH + 'autoreconf'
else:
    if os.getenv('AUTORECONF'):
        UTILS['autoreconf'] = os.getenv('AUTORECONF')
    else:
        UTILS['autoreconf'] = 'autoreconf'

# You can also set the variable AUTOHEADER individually
if AUTOCONFPATH:
    UTILS['autoheader'] = AUTOCONFPATH + 'autoheader'
else:
    if os.getenv('AUTOHEADER'):
        UTILS['autoheader'] = os.getenv('AUTOHEADER')
    else:
        UTILS['autoheader'] = 'autoheader'

# You can also set the variable AUTOMAKE individually
if AUTOMAKEPATH:
    UTILS['automake'] = AUTOMAKEPATH + 'automake'
else:
    if os.getenv('AUTOMAKE'):
        UTILS['automake'] = os.getenv('AUTOMAKE')
    else:
        UTILS['automake'] = 'automake'

# You can also set the variable ACLOCAL individually
if AUTOMAKEPATH:
    UTILS['aclocal'] = AUTOMAKEPATH + 'aclocal'
else:
    if os.getenv('ACLOCAL'):
        UTILS['aclocal'] = os.getenv('ACLOCAL')
    else:
        UTILS['aclocal'] = 'aclocal'

# You can also set the variable AUTOPOINT individually
if GETTEXTPATH:
    UTILS['autopoint'] = GETTEXTPATH + 'autopoint'
else:
    if os.getenv('AUTOPOINT'):
        UTILS['autopoint'] = os.getenv('AUTOPOINT')
    else:
        UTILS['autopoint'] = 'autopoint'

# You can also set the variable LIBTOOLIZE individually
if LIBTOOLPATH:
    UTILS['libtoolize'] = LIBTOOLPATH + 'libtoolize'
else:
    if os.getenv('LIBTOOLIZE'):
        UTILS['libtoolize'] = os.getenv('LIBTOOLIZE')
    else:
        UTILS['libtoolize'] = 'libtoolize'

# You can also set the variable MAKE
if os.getenv('MAKE'):
    UTILS['make'] = os.getenv('MAKE')
else:
    UTILS['make'] = 'make'


#===============================================================================
# Define global functions
#===============================================================================

def force_output() -> None:
    '''This function is to be invoked before invoking external programs.
    It initiates bringing the the contents of process-internal output buffers
    to their respective destinations.'''
    sys.stdout.flush()
    sys.stderr.flush()


def execute(args: list[str], verbose: int) -> None:
    '''Execute the given shell command.'''
    if verbose >= 0:
        print('executing %s' % ' '.join(args), flush=True)
        try:  # Try to run
            retcode = sp.call(args)
        except Exception as error:
            sys.stderr.write(str(error) + '\n')
            sys.exit(1)
    else:
        # Commands like automake produce output to stderr even when they succeed.
        # Turn this output off if the command succeeds.
        temp = tempfile.mktemp()
        xargs = '%s > %s 2>&1' % (' '.join(args), temp)
        try:  # Try to run
            retcode = sp.call(xargs, shell=True)
        except Exception as error:
            sys.stderr.write(str(error) + '\n')
            sys.exit(1)
        if retcode == 0:
            os.remove(temp)
        else:
            print('executing %s' % ' '.join(args))
            with open(temp, mode='r', newline='\n', encoding='utf-8') as file:
                cmdout = file.read()
            print(cmdout)
            os.remove(temp)
            sys.exit(retcode)


def cleaner(sequence: str | list[str]) -> str | list[str | bool]:
    '''Clean string or list of strings after using regex.'''
    if type(sequence) is str:
        sequence = sequence.replace('[', '')
        sequence = sequence.replace(']', '')
    elif type(sequence) is list:
        sequence = [ value.replace('[', '').replace(']', '')
                     for value in sequence]
        sequence = [ value.replace('(', '').replace(')', '')
                     for value in sequence]
        sequence = [ False if value == 'false' else value
                     for value in sequence ]
        sequence = [ True if value == 'true' else value
                     for value in sequence ]
        sequence = [ value.strip()
                     if type(value) is str else value
                     for value in sequence ]
    return sequence


def joinpath(head: str, *tail: str) -> str:
    '''Join two or more pathname components, inserting '/' as needed. If any
    component is an absolute path, all previous path components will be
    discarded.
    This function also replaces SUBDIR/../ with empty; therefore it is not
    suitable when some of the pathname components use Makefile variables
    such as '$(srcdir)'.'''
    return os.path.normpath(os.path.join(head, *tail))


def relativize(dir1: str, dir2: str) -> str:
    '''Compute a relative pathname reldir such that dir1/reldir = dir2.
    dir1 and dir2 must be relative pathnames.'''
    dir0 = os.getcwd()
    while dir1:
        dir1 = '%s%s' % (os.path.normpath(dir1), os.path.sep)
        dir2 = '%s%s' % (os.path.normpath(dir2), os.path.sep)
        first = dir1[:dir1.find(os.path.sep)]
        if first != '.':
            if first == '..':
                dir2 = joinpath(os.path.basename(dir0), dir2)
                dir0 = os.path.dirname(dir0)
            else:  # if first != '..'
                # Get first component of dir2
                first2 = dir2[:dir2.find(os.path.sep)]
                if first == first2:
                    dir2 = dir2[dir2.find(os.path.sep) + 1:]
                else:  # if first != first2
                    dir2 = joinpath('..', dir2)
                dir0 = joinpath(dir0, first)
        dir1 = dir1[dir1.find(os.path.sep) + 1:]
    result = os.path.normpath(dir2)
    return result


def relconcat(dir1: str, dir2: str) -> str:
    '''Compute a relative pathname dir1/dir2, with obvious simplifications.
    dir1 and dir2 must be relative pathnames.
    dir2 is considered to be relative to dir1.'''
    return os.path.normpath(os.path.join(dir1, dir2))


def ensure_writable(dest: str) -> None:
    '''Ensure that the file dest is writable.'''
    # os.stat throws FileNotFoundError error but we assume it exists.
    st = os.stat(dest)
    if not (st.st_mode & stat.S_IWUSR):
        os.chmod(dest, st.st_mode | stat.S_IWUSR)


def relinverse(dir: str) -> str:
    '''Compute the inverse of dir. Namely, a relative pathname consisting only
    of '..' components, such that dir/relinverse = '.'.
    dir must be a relative pathname.'''
    if False:
        # This should work too.
        return relativize(dir, '.')
    else:
        inverse = ''
        for component in dir.split('/'):
            if component != '':
                inverse += '../'
        return os.path.normpath(inverse)


def copyfile(src: str, dest: str) -> None:
    '''Copy file src to file dest. Like shutil.copy, but ignore errors e.g. on
    VFAT file systems.'''
    shutil.copyfile(src, dest)
    try:
        shutil.copymode(src, dest)
    except PermissionError:
        pass


def copyfile2(src: str, dest: str) -> None:
    '''Copy file src to file dest, preserving modification time. Like
    shutil.copy2, but ignore errors e.g. on VFAT file systems. This function
    is to be used for backup files.'''
    shutil.copyfile(src, dest)
    try:
        shutil.copystat(src, dest)
    except PermissionError:
        pass


def movefile(src: str, dest: str) -> None:
    '''Move/rename file src to file dest. Like shutil.move, but gracefully
    handle common errors.'''
    try:
        shutil.move(src, dest)
    except PermissionError:
        # shutil.move invokes os.rename, catches the resulting OSError for
        # errno=EXDEV, attempts a copy instead, and encounters a PermissionError
        # while doing that.
        copyfile2(src, dest)
        os.remove(src)


def symlink_relative(src: str, dest: str) -> None:
    '''Like ln -s, except use cp -p if ln -s fails.
    src is either absolute or relative to the directory of dest.'''
    try:
        os.symlink(src, dest)
    except PermissionError:
        sys.stderr.write('%s: ln -s failed; falling back on cp -p\n' % APP['name'])
        if os.path.isabs(src):
            # src is absolute.
            cp_src = src
        else:
            # src is relative to the directory of dest.
            last_slash = dest.rfind('/')
            if last_slash >= 0:
                cp_src = joinpath(dest[0:last_slash-1], src)
            else:
                cp_src = src
        copyfile2(cp_src, dest)
        ensure_writable(dest)


def as_link_value_at_dest(src: str, dest: str) -> str:
    '''Compute the symbolic link value to place at dest, such that the
    resulting symbolic link points to src. src is given relative to the
    current directory (or absolute).'''
    if type(src) is not str:
        raise TypeError('src must be a string, not %s' % (type(src).__name__))
    if type(dest) is not str:
        raise TypeError('dest must be a string, not %s' % (type(dest).__name__))
    if os.path.isabs(src):
        return src
    else:  # if src is not absolute
        if os.path.isabs(dest):
            return joinpath(os.getcwd(), src)
        else:  # if dest is not absolute
            destdir = os.path.dirname(dest)
            if not destdir:
                destdir = '.'
            return relativize(destdir, src)


def link_relative(src: str, dest: str) -> None:
    '''Like ln -s, except that src is given relative to the current directory
    (or absolute), not given relative to the directory of dest.'''
    if type(src) is not str:
        raise TypeError('src must be a string, not %s' % (type(src).__name__))
    if type(dest) is not str:
        raise TypeError('dest must be a string, not %s' % (type(dest).__name__))
    link_value = as_link_value_at_dest(src, dest)
    symlink_relative(link_value, dest)


def link_if_changed(src: str, dest: str) -> None:
    '''Create a symlink, but avoids munging timestamps if the link is correct.'''
    link_value = as_link_value_at_dest(src, dest)
    if not (os.path.islink(dest) and os.readlink(dest) == link_value):
        try:
            os.remove(dest)
        except FileNotFoundError:
            pass
        # Equivalent to link_relative(src, dest):
        symlink_relative(link_value, dest)


def hardlink(src: str, dest: str) -> None:
    '''Like ln, except use cp -p if ln fails.
    src is either absolute or relative to the directory of dest.'''
    try:
        os.link(src, dest)
    except PermissionError:
        sys.stderr.write('%s: ln failed; falling back on cp -p\n' % APP['name'])
        if os.path.isabs(src):
            # src is absolute.
            cp_src = src
        else:
            # src is relative to the directory of dest.
            last_slash = dest.rfind('/')
            if last_slash >= 0:
                cp_src = joinpath(dest[0: last_slash - 1], src)
            else:
                cp_src = src
        copyfile2(cp_src, dest)
        ensure_writable(dest)


def rmtree(dest: str) -> None:
    '''Removes the file or directory tree at dest, if it exists.'''
    # These two implementations are nearly equivalent.
    # Speed: 'rm -rf' can be a little faster.
    # Exceptions: shutil.rmtree raises Python exceptions, e.g. PermissionError.
    if True:
        sp.run(['rm', '-rf', dest], shell=False)
    else:
        try:
            shutil.rmtree(dest)
        except FileNotFoundError:
            pass


def filter_filelist(separator: str, filelist: list[str], prefix: str, suffix: str,
                    removed_prefix: str, removed_suffix: str,
                    added_prefix: str = '', added_suffix: str = '') -> str:
    '''Filter the given list of files. Filtering: Only the elements starting with
    prefix and ending with suffix are considered. Processing: removed_prefix
    and removed_suffix are removed from each element, added_prefix and
    added_suffix are added to each element.'''
    listing = []
    for filename in filelist:
        if filename.startswith(prefix) and filename.endswith(suffix):
            pattern = re.compile(r'^%s(.*)%s$'
                                 % (removed_prefix, removed_suffix))
            result = pattern.sub(r'%s\1%s'
                                 % (added_prefix, added_suffix), filename)
            listing.append(result)
    # Return an empty string if no files were matched, else combine them
    # with the given separator.
    if listing:
        result = separator.join(listing)
    else:
        result = ''
    return result


def lines_to_multiline(lines: list[str]) -> str:
    '''Combine the lines to a single string, terminating each line with a
    newline character.'''
    if len(lines) > 0:
        return '\n'.join(lines) + '\n'
    else:
        return ''


def substart(orig: str, repl: str, data: str) -> str:
    '''Replaces the start portion of a string.

    Returns data with orig replaced by repl, but only at the beginning of data.
    Like data.replace(orig,repl), except only at the beginning of data.'''
    result = data
    if data.startswith(orig):
        result = repl + data[len(orig):]
    return result


def subend(orig: str, repl: str, data: str) -> str:
    '''Replaces the end portion of a string.

    Returns data with orig replaced by repl, but only at the end of data.
    Like data.replace(orig,repl), except only at the end of data.'''
    result = data
    if data.endswith(orig):
        result = data[:-len(orig)] + repl
    return result


def remove_trailing_slashes(text: str) -> str:
    '''Remove trailing slashes from a file name, except when the file name
    consists only of slashes.'''
    result = text
    while result.endswith('/'):
        result = result[:-1]
        if result == '':
            result = text
            break
    return result


def combine_lines(text: str) -> str:
    '''Given a multiline string text, join lines by spaces:
    When a line ends in a backslash, remove the backslash and join the next
    line to it, inserting a space between them.'''
    return text.replace('\\\n', ' ')


def combine_lines_matching(pattern: re.Pattern, text: str) -> str:
    '''Given a multiline string text, join lines by spaces, when the first
    such line matches a given RegexObject pattern.
    When a line that matches the pattern ends in a backslash, remove the
    backslash and join the next line to it, inserting a space between them.
    When a line that is the result of such a join ends in a backslash,
    proceed likewise.'''
    outerpos = 0
    match = pattern.search(text, outerpos)
    while match:
        (startpos, pos) = match.span()
        # Look how far the continuation lines extend.
        pos = text.find('\n', pos)
        while pos > 0 and text[pos - 1] == '\\':
            pos = text.find('\n', pos + 1)
        if pos < 0:
            pos = len(text)
        # Perform a combine_lines throughout the continuation lines.
        partdone = text[:startpos] + combine_lines(text[startpos:pos])
        outerpos = len(partdone)
        text = partdone + text[pos:]
        # Next round.
        match = pattern.search(text, outerpos)
    return text


def get_terminfo_string(capability: str) -> str:
    '''Returns the value of a string-type terminfo capability for the current value of $TERM.
    Returns the empty string if not defined.'''
    value = ''
    try:
        value = sp.run(['tput', capability], stdout=sp.PIPE, stderr=sp.DEVNULL).stdout.decode('utf-8')
    except Exception:
        pass
    return value


def bold_escapes() -> tuple[str, str]:
    '''Returns the escape sequences for turning bold-face on and off.'''
    term = os.getenv('TERM', '')
    if term != '' and os.isatty(1):
        if term.startswith('xterm'):
            # Assume xterm compatible escape sequences.
            bold_on = '\033[1m'
            bold_off = '\033[0m'
        else:
            # Use the terminfo capability strings for "bold" and "sgr0".
            if term == 'sun-color' and get_terminfo_string('smso') != get_terminfo_string('rev'):
                # Solaris 11 OmniOS: `tput smso` renders as bold,
                #                    `tput rmso` is the same as `tput sgr0`.
                bold_on = get_terminfo_string('smso')
                bold_off = get_terminfo_string('rmso')
            else:
                bold_on = get_terminfo_string('bold')
                bold_off = get_terminfo_string('sgr0')
            if bold_on == '' or bold_off == '':
                bold_on = ''
                bold_off = ''
    else:
        bold_on = ''
        bold_off = ''
    return (bold_on, bold_off)


__all__ += ['APP', 'DIRS', 'MODES', 'UTILS']



# Copyright (C) 2002-2026 Free Software Foundation, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

# This program is meant for authors or maintainers which want to import
# modules from gnulib into their packages.

# This program assumes a Python version >= 3.7.

# CODING STYLE for this file and its companions:
# Like PEP 8 <https://peps.python.org/pep-0008/>, except
# - Line length is not limited to 79 characters.
# - Line breaking before or after binary operators? Better before, like in GNU.
# - Place line breaks to help reading the code:
#   Avoid line breaks inside expressions, if they can be avoided.
#   Do use line breaks to separate the parts of [ ... for ... ] iterations.
# - String literals: Use 'single-quotes'.
#   Rationale: <https://lists.gnu.org/archive/html/bug-gnulib/2024-02/msg00226.html>
# - Use raw string syntax for regular expression pattern strings and repl
#   strings:
#     re.compile(r'...').sub(r'...', ...)
#     re.sub(r'...', r'...', ...)
#   Rationale: This avoids the need to double backslashes, making the pattern
#   and repl strings easier to understand.
# - Comparison operators:
#   Use == and != for variables which can contain only strings, numbers, lists,
#   and None.
#   Use 'is' and 'is not' for variables which can contain only class instances
#   (e.g. GLModule instances) and None.
#   Rationale: <https://lists.gnu.org/archive/html/bug-gnulib/2024-02/msg00225.html>
# - Avoid conditional expressions as in PEP 308 <https://peps.python.org/pep-0308/>.
#   Rationale: They violate the principle that in conditional code, the
#   condition comes first.
# - Use type hints to document functions and catch programming mistakes.
#   To allow for type annotations and union type specifiers, using the
#   syntax Type1 | Type2, for all versions of Python ≥ 3.7 use the
#   'from __future__ import annotations' import at the top of files.
#   Cf. <https://docs.python.org/3/library/__future__.html>
#   and <https://stackoverflow.com/questions/73879925/>.
# - Prefer the standard collection type hints over those provided in the
#   typing module. The latter are deprecated and may be removed.
#   Cf. <https://peps.python.org/pep-0585/>.
# - Never use the {} literal, because it denotes a dictionary,
#   as opposed to {x}, {x,y}, etc., which denote sets.
# - Use a leading underscore to indicate that functions and instance variables
#   should be considered private. This is a widely accepted convention from
#   PEP 8 <https://peps.python.org/pep-0008/#method-names-and-instance-variables>.

# You can use this command to check the style:
#   $ pycodestyle *.py
# The pycodestyle configuration is found in the file setup.cfg.
# Documentation:
# <https://pycodestyle.pycqa.org/en/latest/intro.html#error-codes>

# You can use this command to check for mistakes:
#   $ pylint *.py
# The pylint configuration is found in the file .pylintrc.
# Note: This configuration is only effective when you invoke pylint from this
# directory.
# Documentation:
# <https://pylint.readthedocs.io/en/latest/user_guide/messages/messages_overview.html>

from __future__ import annotations

#===============================================================================
# Define global imports
#===============================================================================
import os
import re
import sys
import random
import argparse
import subprocess as sp
import shlex
import tempfile
from pygnulib.constants import (
    APP,
    DIRS,
    TESTS,
    UTILS,
    MODES,
    ENCS,
    joinpath,
    lines_to_multiline,
    ensure_writable,
    copyfile,
    force_output,
    init_DIRS,
    rmtree,
    DEFAULT_AUTOCONF_MINVERSION,
)
from pygnulib.functions import rewrite_file_name
from pygnulib.enums import CopyAction
from pygnulib.GLConfig import GLConfig
from pygnulib.GLError import GLError
from pygnulib.GLFileSystem import GLFileSystem
from pygnulib.GLFileSystem import GLFileAssistant
from pygnulib.GLImport import GLImport
from pygnulib.GLInfo import GLInfo
from pygnulib.GLModuleSystem import GLModuleSystem
from pygnulib.GLTestDir import GLTestDir
from pygnulib.GLTestDir import GLMegaTestDir


#===============================================================================
# Define main part
#===============================================================================
def main(temp_directory: str) -> None:
    info = GLInfo()
    parser = argparse.ArgumentParser(
        prog=APP['name'],
        usage='gnulib-tool.py --help',
        add_help=False)

    # Here we list the options in the order they are listed in the --help output.

    # list
    parser.add_argument('--list',
                        dest='mode_list',
                        default=None,
                        action='store_true')
    # find
    parser.add_argument('--find',
                        dest='mode_find',
                        default=None,
                        action='store_true')
    # import
    parser.add_argument('--import',
                        dest='mode_import',
                        default=None,
                        action='store_true')
    # add-import
    parser.add_argument('--add-import',
                        dest='mode_add_import',
                        default=None,
                        action='store_true')
    # remove-import
    parser.add_argument('--remove-import',
                        dest='mode_remove_import',
                        default=None,
                        action='store_true')
    # update
    parser.add_argument('--update',
                        dest='mode_update',
                        default=None,
                        action='store_true')
    # create-testdir
    parser.add_argument('--create-testdir',
                        dest='mode_create_testdir',
                        default=None,
                        action='store_true')
    # create-megatestdir
    parser.add_argument('--create-megatestdir',
                        dest='mode_create_megatestdir',
                        default=None,
                        action='store_true')
    # test
    parser.add_argument('--test',
                        dest='mode_test',
                        default=None,
                        action='store_true')
    # megatest
    parser.add_argument('--megatest',
                        dest='mode_megatest',
                        default=None,
                        action='store_true')
    # extract-*
    parser.add_argument('--extract-description',
                        dest='mode_xdescription',
                        default=None,
                        action='store_true')
    parser.add_argument('--extract-comment',
                        dest='mode_xcomment',
                        default=None,
                        action='store_true')
    parser.add_argument('--extract-status',
                        dest='mode_xstatus',
                        default=None,
                        action='store_true')
    parser.add_argument('--extract-notice',
                        dest='mode_xnotice',
                        default=None,
                        action='store_true')
    parser.add_argument('--extract-applicability',
                        dest='mode_xapplicability',
                        default=None,
                        action='store_true')
    parser.add_argument('--extract-usability-in-testdir',
                        dest='mode_xusability_in_testdir',
                        default=None,
                        action='store_true')
    parser.add_argument('--extract-filelist',
                        dest='mode_xfilelist',
                        default=None,
                        action='store_true')
    parser.add_argument('--extract-dependencies',
                        dest='mode_xdependencies',
                        default=None,
                        action='store_true')
    parser.add_argument('--extract-recursive-dependencies',
                        dest='mode_xrecursive_dependencies',
                        default=None,
                        action='store_true')
    parser.add_argument('--extract-dependents',
                        dest='mode_xdependents',
                        default=None,
                        action='store_true')
    parser.add_argument('--extract-recursive-dependents',
                        dest='mode_xrecursive_dependents',
                        default=None,
                        action='store_true')
    parser.add_argument('--extract-autoconf-snippet',
                        dest='mode_xautoconf',
                        default=None,
                        action='store_true')
    parser.add_argument('--extract-automake-snippet',
                        dest='mode_xautomake',
                        default=None,
                        action='store_true')
    parser.add_argument('--extract-include-directive',
                        dest='mode_xinclude',
                        default=None,
                        action='store_true')
    parser.add_argument('--extract-link-directive',
                        dest='mode_xlink',
                        default=None,
                        action='store_true')
    parser.add_argument('--extract-recursive-link-directive',
                        dest='mode_xrecursive_link',
                        default=None,
                        action='store_true')
    parser.add_argument('--extract-license',
                        dest='mode_xlicense',
                        default=None,
                        action='store_true')
    parser.add_argument('--extract-maintainer',
                        dest='mode_xmaintainer',
                        default=None,
                        action='store_true')
    parser.add_argument('--extract-tests-module',
                        dest='mode_xtests',
                        default=None,
                        action='store_true')
    # copy-file
    parser.add_argument('--copy-file',
                        dest='mode_copy_file',
                        default=None,
                        action='store_true')
    # help
    parser.add_argument('--help', '--hel', '--he', '--h',
                        dest='help',
                        default=None,
                        action='store_true')
    # version
    parser.add_argument('--version', '--versio', '--versi', '--vers',
                        dest='version',
                        default=None,
                        action='store_true')
    # no-changelog: a no-op for backward compatibility
    parser.add_argument('--no-changelog',
                        dest='changelog',
                        default=None,
                        action='store_false')
    # destdir
    parser.add_argument('--dir',
                        dest='destdir',
                        default=None,
                        nargs=1)
    # localpath
    parser.add_argument('--local-dir',
                        action='append',
                        dest='localpath',
                        default=None,
                        nargs=1)
    # cache-modules: a no-op for backward compatibility
    parser.add_argument('--cache-modules',
                        dest='cache_modules',
                        default=None,
                        action='store_true')
    parser.add_argument('--no-cache-modules',
                        dest='cache_modules',
                        default=None,
                        action='store_false')
    # verbose
    parser.add_argument('--verbose',
                        default=0,
                        action='count')
    # quiet
    parser.add_argument('--quiet',
                        default=0,
                        action='count')
    # dryrun
    parser.add_argument('--dry-run',
                        dest='dryrun',
                        default=None,
                        action='store_true')
    # inctests
    parser.add_argument('--with-tests',
                        dest='inctests',
                        default=None,
                        action='store_true')
    parser.add_argument('--without-tests',
                        dest='inctests',
                        default=None,
                        action='store_false')
    # obsolete
    parser.add_argument('--with-obsolete',
                        dest='obsolete',
                        default=None,
                        action='store_true')
    # c++-tests
    parser.add_argument('--with-c++-tests',
                        dest='inc_cxx_tests',
                        default=None,
                        action='store_true')
    parser.add_argument('--without-c++-tests',
                        dest='excl_cxx_tests',
                        default=None,
                        action='store_true')
    # longrunning-tests
    parser.add_argument('--with-longrunning-tests',
                        dest='inc_longrunning_tests',
                        default=None,
                        action='store_true')
    parser.add_argument('--without-longrunning-tests',
                        dest='excl_longrunning_tests',
                        default=None,
                        action='store_true')
    # privileged-tests
    parser.add_argument('--with-privileged-tests',
                        dest='inc_privileged_tests',
                        default=None,
                        action='store_true')
    parser.add_argument('--without-privileged-tests',
                        dest='excl_privileged_tests',
                        default=None,
                        action='store_true')
    # unportable-tests
    parser.add_argument('--with-unportable-tests',
                        dest='inc_unportable_tests',
                        default=None,
                        action='store_true')
    parser.add_argument('--without-unportable-tests',
                        dest='excl_unportable_tests',
                        default=None,
                        action='store_true')
    # all-tests
    parser.add_argument('--with-all-tests',
                        dest='alltests',
                        default=None,
                        action='store_true')
    # avoids
    parser.add_argument('--avoid',
                        dest='avoids',
                        default=None,
                        action='append',
                        nargs=1)
    # conditional-dependencies
    parser.add_argument('--conditional-dependencies',
                        dest='cond_dependencies',
                        default=None,
                        action='store_true')
    parser.add_argument('--no-conditional-dependencies',
                        dest='cond_dependencies',
                        default=None,
                        action='store_false')
    # libtool
    parser.add_argument('--libtool',
                        dest='libtool',
                        default=None,
                        action='store_true')
    parser.add_argument('--no-libtool',
                        dest='libtool',
                        default=None,
                        action='store_false')
    # libname
    parser.add_argument('--lib',
                        dest='libname',
                        default=None,
                        nargs=1)
    # sourcebase
    parser.add_argument('--source-base',
                        dest='sourcebase',
                        default=None,
                        nargs=1)
    # m4base
    parser.add_argument('--m4-base',
                        dest='m4base',
                        default=None,
                        nargs=1)
    # pobase
    parser.add_argument('--po-base',
                        dest='pobase',
                        default=None,
                        nargs=1)
    # docbase
    parser.add_argument('--doc-base',
                        dest='docbase',
                        default=None,
                        nargs=1)
    # testsbase
    parser.add_argument('--tests-base',
                        dest='testsbase',
                        default=None,
                        nargs=1)
    # auxdir
    parser.add_argument('--aux-dir',
                        dest='auxdir',
                        default=None,
                        nargs=1)
    # lgpl
    parser.add_argument('--lgpl',
                        dest='lgpl',
                        default=None,
                        action='append',
                        choices=['2', '3orGPLv2', '3'],
                        nargs='?')
    # gpl
    parser.add_argument('--gpl',
                        dest='gpl',
                        default=None,
                        choices=['2', '3'],
                        nargs=1)
    # gnu-make
    parser.add_argument('--gnu-make',
                        dest='gnu_make',
                        default=None,
                        action='store_true')
    # makefile-name
    parser.add_argument('--makefile-name',
                        dest='makefile_name',
                        default=None,
                        nargs=1)
    # tests-makefile-name
    parser.add_argument('--tests-makefile-name',
                        dest='tests_makefile_name',
                        default=None,
                        nargs=1)
    # automake-subdir
    parser.add_argument('--automake-subdir',
                        dest='automake_subdir',
                        default=None,
                        action='store_true')
    # automake-subdir-tests
    parser.add_argument('--automake-subdir-tests',
                        dest='automake_subdir_tests',
                        default=None,
                        action='store_true')
    # macro-prefix
    parser.add_argument('--macro-prefix',
                        dest='macro_prefix',
                        default=None,
                        nargs=1)
    # po-domain
    parser.add_argument('--po-domain',
                        dest='podomain',
                        default=None,
                        nargs=1)
    # witness-c-macro
    parser.add_argument('--witness-c-macro',
                        dest='witness_c_macro',
                        default=None,
                        nargs=1)
    # vc-files
    parser.add_argument('--vc-files',
                        dest='vc_files',
                        default=None,
                        action='store_true')
    parser.add_argument('--no-vc-files',
                        dest='vc_files',
                        default=None,
                        action='store_false')
    # two-configures, single-configure
    parser.add_argument('--two-configures',
                        dest='single_configure',
                        default=None,
                        action='store_false')
    parser.add_argument('--single-configure',
                        dest='single_configure',
                        default=None,
                        action='store_true')
    # symlink
    parser.add_argument('-s', '-S', '--symbolic', '--symlink', '--more-symlinks',
                        dest='copymode',
                        default=None,
                        action='store_const', const=CopyAction.Symlink)
    # local-symlink
    parser.add_argument('--local-symlink',
                        dest='lcopymode',
                        default=None,
                        action='store_const', const=CopyAction.Symlink)
    # hardlink
    parser.add_argument('-h', '-H', '--hardlink', '--more-hardlinks',
                        dest='copymode',
                        default=None,
                        action='store_const', const=CopyAction.Hardlink)
    # local-hardlink
    parser.add_argument('--local-hardlink',
                        dest='lcopymode',
                        default=None,
                        action='store_const', const=CopyAction.Hardlink)
    # Undocumented option. Only used for the gnulib-tool test suite.
    parser.add_argument('--gnulib-dir',
                        dest='gnulib_dir',
                        default=None,
                        nargs=1)
    # All other arguments are collected.
    parser.add_argument('non_option_arguments',
                        nargs='*')

    # Parse the given arguments. Don't signal an error if non-option arguments
    # occur between or after options.
    cmdargs, unhandled = parser.parse_known_args()

    # Handle --gnulib-dir and finalize DIRS.
    gnulib_dir = cmdargs.gnulib_dir
    if gnulib_dir != None:
        gnulib_dir = os.path.abspath(gnulib_dir[0])
    else:
        gnulib_dir = APP['root']
    init_DIRS(gnulib_dir)

    # Handle --help and --version, ignoring all other options.
    if cmdargs.help != None:
        print(info.usage())
        sys.exit(0)
    if cmdargs.version != None:
        version = info.version()
        if version != '':
            version = ' ' + version
        message = '''gnulib-tool (%s %s)%s\n%s\n%s\n\nWritten by %s.''' \
            % (info.package(), info.date(), version, info.copyright(),
               info.license(), info.authors())
        print(message)
        sys.exit(0)

    # Report unhandled arguments.
    for arg in unhandled:
        if arg.startswith('-'):
            message = '%s: Unrecognized option \'%s\'.\n' % (APP['name'], arg)
            message += 'Try \'gnulib-tool --help\' for more information.\n'
            sys.stderr.write(message)
            sys.exit(1)
    # By now, all unhandled arguments were non-options.
    cmdargs.non_option_arguments += unhandled

    # Determine when user tries to combine modes.
    args = [
        cmdargs.mode_list,
        cmdargs.mode_find,
        cmdargs.mode_import,
        cmdargs.mode_add_import,
        cmdargs.mode_remove_import,
        cmdargs.mode_update,
        cmdargs.mode_create_testdir,
        cmdargs.mode_create_megatestdir,
        cmdargs.mode_test,
        cmdargs.mode_megatest,
        cmdargs.mode_xdescription,
        cmdargs.mode_xcomment,
        cmdargs.mode_xstatus,
        cmdargs.mode_xnotice,
        cmdargs.mode_xapplicability,
        cmdargs.mode_xusability_in_testdir,
        cmdargs.mode_xfilelist,
        cmdargs.mode_xdependencies,
        cmdargs.mode_xrecursive_dependencies,
        cmdargs.mode_xdependents,
        cmdargs.mode_xrecursive_dependents,
        cmdargs.mode_xautoconf,
        cmdargs.mode_xautomake,
        cmdargs.mode_xinclude,
        cmdargs.mode_xlink,
        cmdargs.mode_xlicense,
        cmdargs.mode_xmaintainer,
        cmdargs.mode_xtests,
        cmdargs.mode_copy_file,
    ]
    overflow = [ arg
                 for arg in args
                 if arg != None ]
    if len(overflow) > 1:
        message = '%s: Unable to combine different modes of work.\n' % APP['name']
        message += 'Try \'gnulib-tool --help\' for more information.\n'
        sys.stderr.write(message)
        sys.exit(1)

    # Determine selected mode.
    mode = None
    modules = None
    files = None
    if cmdargs.mode_list != None:
        mode = 'list'
    if cmdargs.mode_find != None:
        mode = 'find'
        files = list(cmdargs.non_option_arguments)
    if cmdargs.mode_import != None:
        mode = 'import'
        modules = list(cmdargs.non_option_arguments)
    if cmdargs.mode_add_import != None:
        mode = 'add-import'
        modules = list(cmdargs.non_option_arguments)
    if cmdargs.mode_remove_import != None:
        mode = 'remove-import'
        modules = list(cmdargs.non_option_arguments)
    if cmdargs.mode_update != None:
        mode = 'update'
        if len(cmdargs.non_option_arguments) > 0:
            message = '%s: too many arguments in \'update\' mode\n'
            message += 'Try \'gnulib-tool --help\' for more information.\n'
            message += 'If you really want to modify the gnulib configuration of your project,\n'
            message += 'you need to use \'gnulib-tool --import\' - at your own risk!\n'
            sys.stderr.write(message)
            sys.exit(1)
    if cmdargs.mode_create_testdir != None:
        mode = 'create-testdir'
        modules = list(cmdargs.non_option_arguments)
    if cmdargs.mode_create_megatestdir != None:
        mode = 'create-megatestdir'
        modules = list(cmdargs.non_option_arguments)
    if cmdargs.mode_test != None:
        mode = 'test'
        modules = list(cmdargs.non_option_arguments)
    if cmdargs.mode_megatest != None:
        mode = 'megatest'
        modules = list(cmdargs.non_option_arguments)
    if cmdargs.mode_xdescription != None:
        mode = 'extract-description'
        modules = list(cmdargs.non_option_arguments)
    if cmdargs.mode_xcomment != None:
        mode = 'extract-comment'
        modules = list(cmdargs.non_option_arguments)
    if cmdargs.mode_xstatus != None:
        mode = 'extract-status'
        modules = list(cmdargs.non_option_arguments)
    if cmdargs.mode_xnotice != None:
        mode = 'extract-notice'
        modules = list(cmdargs.non_option_arguments)
    if cmdargs.mode_xapplicability != None:
        mode = 'extract-applicability'
        modules = list(cmdargs.non_option_arguments)
    if cmdargs.mode_xusability_in_testdir != None:
        mode = 'extract-usability-in-testdir'
        modules = list(cmdargs.non_option_arguments)
    if cmdargs.mode_xfilelist != None:
        mode = 'extract-filelist'
        modules = list(cmdargs.non_option_arguments)
    if cmdargs.mode_xautoconf != None:
        mode = 'extract-autoconf-snippet'
        modules = list(cmdargs.non_option_arguments)
    if cmdargs.mode_xautomake != None:
        mode = 'extract-automake-snippet'
        modules = list(cmdargs.non_option_arguments)
    if cmdargs.mode_xdependencies != None:
        mode = 'extract-dependencies'
        modules = list(cmdargs.non_option_arguments)
    if cmdargs.mode_xrecursive_dependencies != None:
        mode = 'extract-recursive-dependencies'
        modules = list(cmdargs.non_option_arguments)
    if cmdargs.mode_xdependents != None:
        mode = 'extract-dependents'
        modules = list(cmdargs.non_option_arguments)
    if cmdargs.mode_xrecursive_dependents != None:
        mode = 'extract-recursive-dependents'
        modules = list(cmdargs.non_option_arguments)
    if cmdargs.mode_xinclude != None:
        mode = 'extract-include-directive'
        modules = list(cmdargs.non_option_arguments)
    if cmdargs.mode_xlink != None:
        mode = 'extract-link-directive'
        modules = list(cmdargs.non_option_arguments)
    if cmdargs.mode_xrecursive_link != None:
        mode = 'extract-recursive-link-directive'
        modules = list(cmdargs.non_option_arguments)
    if cmdargs.mode_xlicense != None:
        mode = 'extract-license'
        modules = list(cmdargs.non_option_arguments)
    if cmdargs.mode_xmaintainer != None:
        mode = 'extract-maintainer'
        modules = list(cmdargs.non_option_arguments)
    if cmdargs.mode_xtests != None:
        mode = 'extract-tests-module'
        modules = list(cmdargs.non_option_arguments)
    if cmdargs.mode_copy_file != None:
        mode = 'copy-file'
        if len(cmdargs.non_option_arguments) < 1 or len(cmdargs.non_option_arguments) > 2:
            message = '%s: *** ' % APP['name']
            message += 'invalid number of arguments for --%s\n' % mode
            message += 'Try \'gnulib-tool --help\' for more information.\n'
            message += '%s: *** Stop.\n' % APP['name']
            sys.stderr.write(message)
            sys.exit(1)
        files = list(cmdargs.non_option_arguments)

    if ((mode in ['import', 'add-import', 'remove-import']
         and (cmdargs.excl_cxx_tests or cmdargs.excl_longrunning_tests
              or cmdargs.excl_privileged_tests or cmdargs.excl_unportable_tests
              or cmdargs.single_configure != None))
        or (mode == 'update'
            and (cmdargs.localpath != None or cmdargs.libname != None
                 or cmdargs.sourcebase != None or cmdargs.m4base != None
                 or cmdargs.pobase != None or cmdargs.docbase != None
                 or cmdargs.testsbase != None or cmdargs.auxdir != None
                 or cmdargs.inctests != None or cmdargs.obsolete != None
                 or cmdargs.inc_cxx_tests != None
                 or cmdargs.inc_longrunning_tests != None
                 or cmdargs.inc_privileged_tests != None
                 or cmdargs.inc_unportable_tests != None
                 or cmdargs.alltests != None
                 or cmdargs.excl_cxx_tests != None
                 or cmdargs.excl_longrunning_tests != None
                 or cmdargs.excl_privileged_tests != None
                 or cmdargs.excl_unportable_tests != None
                 or cmdargs.avoids != None or cmdargs.lgpl != None
                 or cmdargs.gpl != None or cmdargs.makefile_name != None
                 or cmdargs.tests_makefile_name != None
                 or cmdargs.automake_subdir != None
                 or cmdargs.automake_subdir_tests != None
                 or cmdargs.macro_prefix != None or cmdargs.podomain != None
                 or cmdargs.witness_c_macro != None or cmdargs.vc_files != None))):
        message = '%s: *** ' % APP['name']
        message += 'invalid options for --%s mode\n' % mode
        message += 'Try \'gnulib-tool --help\' for more information.\n'
        message += '%s: *** Stop.\n' % APP['name']
        sys.stderr.write(message)
        sys.exit(1)
    if cmdargs.pobase != None and cmdargs.podomain == None:
        message = '%s: *** ' % APP['name']
        message += 'together with --po-base, you need to specify --po-domain\n'
        message += 'Try \'gnulib-tool --help\' for more information.\n'
        message += '%s: *** Stop.\n' % APP['name']
        sys.stderr.write(message)
        sys.exit(1)
    if cmdargs.lgpl != None and cmdargs.gpl != None:
        message = '%s: *** ' % APP['name']
        message += 'invalid options: cannot specify both --lgpl and --gpl\n'
        message += 'Try \'gnulib-tool --help\' for more information.\n'
        message += '%s: *** Stop.\n' % APP['name']
        sys.stderr.write(message)
        sys.exit(1)
    if cmdargs.pobase == None and cmdargs.podomain != None:
        message = 'gnulib-tool: warning: --po-domain has no effect without a --po-base option\n'
        sys.stderr.write(message)
    if mode != None and 'test' in mode and cmdargs.gnu_make:
        message = '%s: --gnu-make not supported when including tests\n' % APP['name']
        sys.stderr.write(message)
        sys.exit(1)

    # Determine specific settings.
    destdir = cmdargs.destdir
    if destdir != None:
        destdir = cmdargs.destdir[0]
    localpath = cmdargs.localpath
    if localpath != None:
        localpath = [ dir
                      for list1 in localpath
                      for dir in list1 ]
    libname = cmdargs.libname
    if libname != None:
        libname = cmdargs.libname[0]
    auxdir = cmdargs.auxdir
    if auxdir != None:
        auxdir = cmdargs.auxdir[0]
    sourcebase = cmdargs.sourcebase
    if sourcebase != None:
        sourcebase = cmdargs.sourcebase[0]
    m4base = cmdargs.m4base
    if m4base != None:
        m4base = cmdargs.m4base[0]
    pobase = cmdargs.pobase
    if pobase != None:
        pobase = cmdargs.pobase[0]
    docbase = cmdargs.docbase
    if docbase != None:
        docbase = cmdargs.docbase[0]
    testsbase = cmdargs.testsbase
    if testsbase != None:
        testsbase = cmdargs.testsbase[0]
    dryrun = cmdargs.dryrun
    verbose = -cmdargs.quiet + cmdargs.verbose
    incobsolete = cmdargs.obsolete
    inctests = cmdargs.inctests
    # Canonicalize the inctests variable.
    if inctests == None:
        if mode in ['import', 'add-import', 'remove-import', 'update']:
            inctests = False
        elif mode in ['create-testdir', 'create-megatestdir', 'test', 'megatest']:
            inctests = True
    incl_test_categories = []
    if inctests:
        incl_test_categories.append(TESTS['tests'])
    if cmdargs.inc_cxx_tests:
        incl_test_categories.append(TESTS['cxx-tests'])
    if cmdargs.inc_longrunning_tests:
        incl_test_categories.append(TESTS['longrunning-tests'])
    if cmdargs.inc_privileged_tests:
        incl_test_categories.append(TESTS['privileged-tests'])
    if cmdargs.inc_unportable_tests:
        incl_test_categories.append(TESTS['unportable-tests'])
    if cmdargs.alltests:
        incl_test_categories.append(TESTS['all-tests'])
    excl_test_categories = []
    if cmdargs.excl_cxx_tests:
        excl_test_categories.append(TESTS['cxx-tests'])
    if cmdargs.excl_longrunning_tests:
        excl_test_categories.append(TESTS['longrunning-tests'])
    if cmdargs.excl_privileged_tests:
        excl_test_categories.append(TESTS['privileged-tests'])
    if cmdargs.excl_unportable_tests:
        excl_test_categories.append(TESTS['unportable-tests'])
    lgpl = cmdargs.lgpl
    if lgpl != None:
        lgpl = lgpl[-1]
        if lgpl == None:
            lgpl = True
    gpl = cmdargs.gpl
    if gpl != None:
        gpl = gpl[0]
    cond_dependencies = cmdargs.cond_dependencies
    libtool = cmdargs.libtool
    gnu_make = cmdargs.gnu_make == True
    makefile_name = cmdargs.makefile_name
    if makefile_name != None:
        makefile_name = makefile_name[0]
    tests_makefile_name = cmdargs.tests_makefile_name
    if tests_makefile_name != None:
        tests_makefile_name = tests_makefile_name[0]
    automake_subdir = cmdargs.automake_subdir == True
    automake_subdir_tests = cmdargs.automake_subdir_tests == True
    macro_prefix = cmdargs.macro_prefix
    if macro_prefix != None:
        macro_prefix = macro_prefix[0]
    podomain = cmdargs.podomain
    if podomain != None:
        podomain = podomain[0]
    witness_c_macro = cmdargs.witness_c_macro
    if witness_c_macro != None:
        witness_c_macro = witness_c_macro[0]
    vc_files = cmdargs.vc_files
    avoids = cmdargs.avoids
    if avoids != None:
        avoids = [ module
                   for list1 in avoids
                   for module in list1 ]
    copymode = cmdargs.copymode
    lcopymode = cmdargs.lcopymode
    single_configure = cmdargs.single_configure
    # Canonicalize the single_configure variable.
    if single_configure == None:
        single_configure = True

    # Create pygnulib configuration.
    config = GLConfig(
        tempdir=temp_directory,
        destdir=destdir,
        localpath=localpath,
        m4base=m4base,
        auxdir=auxdir,
        modules=modules,
        avoids=avoids,
        sourcebase=sourcebase,
        pobase=pobase,
        docbase=docbase,
        testsbase=testsbase,
        incobsolete=incobsolete,
        incl_test_categories=incl_test_categories,
        excl_test_categories=excl_test_categories,
        libname=libname,
        lgpl=lgpl,
        gpl=gpl,
        gnu_make=gnu_make,
        makefile_name=makefile_name,
        tests_makefile_name=tests_makefile_name,
        automake_subdir=automake_subdir,
        automake_subdir_tests=automake_subdir_tests,
        libtool=libtool,
        conddeps=cond_dependencies,
        macro_prefix=macro_prefix,
        podomain=podomain,
        witness_c_macro=witness_c_macro,
        vc_files=vc_files,
        copymode=copymode,
        lcopymode=lcopymode,
        single_configure=single_configure,
        verbose=verbose,
        dryrun=dryrun,
    )

    # Work in the given mode.
    if mode == 'list':
        modulesystem = GLModuleSystem(config)
        listing = modulesystem.list()
        result = lines_to_multiline(listing)
        print(result, end='')

    elif mode == 'find':
        modulesystem = GLModuleSystem(config)
        for filename in files:
            if (os.path.isfile(joinpath(DIRS['root'], filename))
                    or (localpath != None
                        and any([ os.path.isfile(joinpath(localdir, filename))
                                  for localdir in localpath ]))):
                # Convert the file name to a POSIX basic regex.
                # Needs to handle . [ \ * ^ $.
                filename_regex = filename.replace('\\', '\\\\').replace('[', '\\[').replace('^', '\\^')
                filename_regex = re.compile(r'([.*$])').sub(r'[\1]', filename_regex)
                filename_line_regex = '^' + filename_regex + '$'
                # Read module candidates from gnulib root directory.
                command = "find modules -type f -print | xargs -n 100 grep -l %s /dev/null | sed -e 's,^modules/,,'" % shlex.quote(filename_line_regex)
                with sp.Popen(command, shell=True, cwd=DIRS['root'], stdout=sp.PIPE) as proc:
                    result = proc.stdout.read().decode('UTF-8')
                # Read module candidates from local directories.
                if localpath != None and len(localpath) > 0:
                    command = "find modules -type f -print | xargs -n 100 grep -l %s /dev/null | sed -e 's,^modules/,,' -e 's,\\.diff$,,'" % shlex.quote(filename_line_regex)
                    for localdir in localpath:
                        with sp.Popen(command, shell=True, cwd=localdir, stdout=sp.PIPE) as proc:
                            result += proc.stdout.read().decode('UTF-8')
                listing = [ line
                            for line in result.split('\n')
                            if line.strip() ]
                # Remove modules/ prefix from each file name.
                pattern = re.compile(r'^modules/')
                listing = [ pattern.sub('', line)
                            for line in listing ]
                # Filter out undesired file names.
                listing = [ line
                            for line in listing
                            if modulesystem.file_is_module(line) ]
                candidates = sorted(set(listing))
                for name in candidates:
                    module = modulesystem.find(name)
                    if module:  # Ignore module candidates that don't actually exist.
                        if module.getFiles():
                            print(name)
            else:
                message = 'gnulib-tool: warning: file %s does not exist\n' % filename
                sys.stderr.write(message)

    elif mode in ['import', 'add-import', 'remove-import', 'update']:
        mode = MODES[mode]
        if not destdir:
            destdir = '.'
        config.setDestDir(destdir)

        # Prefer configure.ac but also look for configure.in.
        # NOTE: Use os.path.join so the leading './' is not removed. This
        # is to make the gnulib-tool test suite happy.
        if os.path.isfile(os.path.join(destdir, 'configure.ac')):
            configure_ac = os.path.join(destdir, 'configure.ac')
        elif os.path.isfile(os.path.join(destdir, 'configure.in')):
            configure_ac = os.path.join(destdir, 'configure.in')
        else:
            raise GLError(3, os.path.join(destdir, 'configure.ac'))

        # Save the Autoconf file path for the rest of the import.
        config.setAutoconfFile(configure_ac)

        # Analyze configure.ac.
        with open(configure_ac, mode='r', newline='\n', encoding='utf-8') as file:
            configure_ac_data = file.read()

        guessed_m4dirs = []
        pattern = re.compile(r'^.*AC_CONFIG_MACRO_DIRS?\([\[ ]*([^]"$`\\)]*).*$', re.MULTILINE)
        match = pattern.findall(configure_ac_data)
        # Append the match to guessed_m4dirs.
        if match:
            guessed_m4dirs += match

        if mode == MODES['import']:
            # Set variables to default values.
            if not sourcebase:
                sourcebase = 'lib'
            if not m4base:
                m4base = 'm4'
            if not docbase:
                docbase = 'doc'
            if not testsbase:
                testsbase = 'tests'
            if not macro_prefix:
                macro_prefix = 'gl'
            config.setSourceBase(sourcebase)
            config.setM4Base(m4base)
            config.setDocBase(docbase)
            config.setTestsBase(testsbase)
            config.setMacroPrefix(macro_prefix)

            # Perform GLImport actions.
            importer = GLImport(config, mode, guessed_m4dirs)
            filetable, transformers = importer.prepare()
            importer.execute(filetable, transformers)

        else:  # if mode != MODE['--import']
            if m4base:
                # Apply func_import to a particular gnulib directory.
                # Any number of additional modules can be given.
                if not os.path.isfile(joinpath(destdir, m4base, 'gnulib-cache.m4')):
                    # First use of gnulib in the given m4base.
                    if not sourcebase:
                        sourcebase = 'lib'
                    if not docbase:
                        docbase = 'doc'
                    if not testsbase:
                        testsbase = 'tests'
                    if not macro_prefix:
                        macro_prefix = 'gl'
                    config.setSourceBase(sourcebase)
                    config.setM4Base(m4base)
                    config.setDocBase(docbase)
                    config.setTestsBase(testsbase)
                    config.setMacroPrefix(macro_prefix)
                # Perform GLImport actions.
                importer = GLImport(config, mode, guessed_m4dirs)
                filetable, transformers = importer.prepare()
                importer.execute(filetable, transformers)
            else:  # if not m4base
                # Apply func_import to all gnulib directories.
                # To get this list of directories, look at Makefile.am. (Not at
                # configure, because it may be omitted from version control. Also,
                # don't run "find $destdir -name gnulib-cache.m4", as it might be
                # too expensive.)
                m4dirs = []
                dirisnext = False
                filepath = joinpath(destdir, 'Makefile.am')
                if os.path.isfile(filepath):
                    with open(filepath, mode='r', newline='\n', encoding='utf-8') as file:
                        data = file.read()
                    pattern = re.compile(r'^ACLOCAL_AMFLAGS[\t ]*=[\t ]*(.+?)[\t ]*(?:#|$)', re.MULTILINE)
                    match = re.search(pattern, data)
                    if match:
                        aclocal_amflags = match.group(1).split()
                    else:
                        aclocal_amflags = []
                    for aclocal_amflag in aclocal_amflags:
                        if dirisnext:
                            # Ignore absolute directory pathnames, like /usr/local/share/aclocal.
                            if not os.path.isabs(aclocal_amflag):
                                if os.path.isfile(joinpath(destdir, aclocal_amflag, 'gnulib-cache.m4')):
                                    m4dirs.append(aclocal_amflag)
                            dirisnext = False
                        else:  # if not dirisnext
                            if aclocal_amflag == '-I':
                                dirisnext = True
                            else:  # if aclocal_amflag != '-I'
                                dirisnext = False
                    for arg in guessed_m4dirs:
                        # Ignore absolute directory pathnames, like /usr/local/share/aclocal.
                        if not os.path.isabs(arg):
                            if os.path.isfile(joinpath(destdir, arg, 'gnulib-cache.m4')):
                                m4dirs.append(arg)
                else:  # if not os.path.isfile(filepath)
                    # No Makefile.am! Oh well. Look at the last generated aclocal.m4.
                    filepath = joinpath(destdir, 'aclocal.m4')
                    if os.path.isfile(filepath):
                        pattern = re.compile(r'm4_include\(\[(.*?)]\)')
                        with open(filepath, mode='r', newline='\n', encoding='utf-8') as file:
                            m4dirs = pattern.findall(file.read())
                        m4dirs = [ os.path.dirname(m4dir)
                                   for m4dir in m4dirs ]
                        m4dirs = [ m4dir
                                   for m4dir in m4dirs
                                   if os.path.isfile(joinpath(destdir, m4dir, 'gnulib-cache.m4')) ]
                        m4dirs = sorted(set(m4dirs))
                if len(m4dirs) == 0:
                    # First use of gnulib in a package.
                    # Any number of additional modules can be given.
                    if not sourcebase:
                        sourcebase = 'lib'
                    m4base = 'm4'
                    if not docbase:
                        docbase = 'doc'
                    if not testsbase:
                        testsbase = 'tests'
                    if not macro_prefix:
                        macro_prefix = 'gl'
                    config.setSourceBase(sourcebase)
                    config.setM4Base(m4base)
                    config.setDocBase(docbase)
                    config.setTestsBase(testsbase)
                    config.setMacroPrefix(macro_prefix)
                    # Perform GLImport actions.
                    importer = GLImport(config, mode, guessed_m4dirs)
                    filetable, transformers = importer.prepare()
                    importer.execute(filetable, transformers)
                elif len(m4dirs) == 1:
                    # There's only one use of gnulib here. Assume the user means it.
                    # Any number of additional modules can be given.
                    m4base = m4dirs[-1]
                    config.setM4Base(m4base)
                    # Perform GLImport actions.
                    importer = GLImport(config, mode, guessed_m4dirs)
                    filetable, transformers = importer.prepare()
                    importer.execute(filetable, transformers)
                else:  # if len(m4dirs) > 1
                    # No further arguments. Guess the user wants to update all of them.
                    for m4base in m4dirs:
                        config.setM4Base(m4base)
                        # Perform GLImport actions.
                        importer = GLImport(config, mode, guessed_m4dirs)
                        filetable, transformers = importer.prepare()
                        importer.execute(filetable, transformers)

    elif mode == 'create-testdir':
        if not destdir:
            message = '%s: *** ' % APP['name']
            message += 'please specify --dir option\n'
            message += '%s: *** Stop.\n' % APP['name']
            sys.stderr.write(message)
            sys.exit(1)
        if not auxdir:
            auxdir = 'build-aux'
        config.setAuxDir(auxdir)
        testdir = GLTestDir(config, destdir)
        testdir.execute()

    elif mode == 'create-megatestdir':
        if not destdir:
            message = '%s: *** ' % APP['name']
            message += 'please specify --dir option\n'
            message += '%s: *** Stop.\n' % APP['name']
            sys.stderr.write(message)
            sys.exit(1)
        if not auxdir:
            auxdir = 'build-aux'
        config.setAuxDir(auxdir)
        testdir = GLMegaTestDir(config, destdir)
        testdir.execute()

    elif mode == 'test':
        if not destdir:
            destdir = 'testdir%04d' % random.randrange(0, 9999)
        if not auxdir:
            auxdir = 'build-aux'
        config.setAuxDir(auxdir)
        testdir = GLTestDir(config, destdir)
        testdir.execute()
        force_output()
        os.chdir(destdir)
        os.mkdir('build')
        os.chdir('build')
        try:  # Try to execute commands
            sp.run(['../configure'], check=True)
            sp.run([UTILS['make']], check=True)
            sp.run([UTILS['make'], 'check'], check=True)
            sp.run([UTILS['make'], 'distclean'], check=True)
        except Exception:
            sys.exit(1)
        args = ['find', '.', '-type', 'f', '-print']
        remaining = sp.check_output(args).decode(ENCS['shell'])
        lines = [ line.strip()
                  for line in remaining.split('\n')
                  if line.strip() ]
        remaining = ' '.join(lines)
        if remaining:
            message = 'Remaining files: %s\n' % remaining
            message += 'gnulib-tool: *** Stop.\n'
            sys.stderr.write(message)
            sys.exit(1)
        os.chdir('../..')
        rmtree(destdir)

    elif mode == 'megatest':
        if not destdir:
            destdir = 'testdir%04d' % random.randrange(0, 9999)
        if not auxdir:
            auxdir = 'build-aux'
        config.setAuxDir(auxdir)
        testdir = GLMegaTestDir(config, destdir)
        testdir.execute()
        force_output()
        os.chdir(destdir)
        os.mkdir('build')
        os.chdir('build')
        try:  # Try to execute commands
            sp.run(['../configure'], check=True)
            sp.run([UTILS['make']], check=True)
            sp.run([UTILS['make'], 'check'], check=True)
            sp.run([UTILS['make'], 'distclean'], check=True)
        except Exception:
            sys.exit(1)
        args = ['find', '.', '-type', 'f', '-print']
        remaining = sp.check_output(args).decode(ENCS['shell'])
        lines = [ line.strip()
                  for line in remaining.split('\n')
                  if line.strip() ]
        remaining = ' '.join(lines)
        if remaining:
            message = 'Remaining files: %s\n' % remaining
            message += 'gnulib-tool: *** Stop.\n'
            sys.stderr.write(message)
            sys.exit(1)
        os.chdir('../..')
        rmtree(destdir)

    elif mode == 'extract-description':
        modulesystem = GLModuleSystem(config)
        for name in modules:
            module = modulesystem.find(name)
            if module:
                sys.stdout.write(module.getDescription())

    elif mode == 'extract-comment':
        modulesystem = GLModuleSystem(config)
        for name in modules:
            module = modulesystem.find(name)
            if module:
                sys.stdout.write(module.getComment())

    elif mode == 'extract-status':
        modulesystem = GLModuleSystem(config)
        for name in modules:
            module = modulesystem.find(name)
            if module:
                sys.stdout.write(module.getStatus())

    elif mode == 'extract-notice':
        modulesystem = GLModuleSystem(config)
        for name in modules:
            module = modulesystem.find(name)
            if module:
                sys.stdout.write(module.getNotice())

    elif mode == 'extract-applicability':
        modulesystem = GLModuleSystem(config)
        for name in modules:
            module = modulesystem.find(name)
            if module:
                print(module.getApplicability())

    elif mode == 'extract-usability-in-testdir':
        modulesystem = GLModuleSystem(config)
        for name in modules:
            module = modulesystem.find(name)
            if module:
                print(module.getUsabilityInTestdir())

    elif mode == 'extract-filelist':
        modulesystem = GLModuleSystem(config)
        for name in modules:
            module = modulesystem.find(name)
            if module:
                files = module.getFiles_Raw()
                files += 'm4/00gnulib.m4\n'
                files += 'm4/zzgnulib.m4\n'
                files += 'm4/gnulib-common.m4\n'
                print(files, end='')

    elif mode == 'extract-dependencies':
        if avoids:
            message = '%s: *** ' % APP['name']
            message += 'cannot combine --avoid and --extract-dependencies\n'
            message += '%s: *** Stop.\n' % APP['name']
            sys.stderr.write(message)
            sys.exit(1)
        modulesystem = GLModuleSystem(config)
        for name in modules:
            module = modulesystem.find(name)
            if module:
                sys.stdout.write(module.getDependencies())

    elif mode == 'extract-recursive-dependencies':
        if avoids:
            message = '%s: *** ' % APP['name']
            message += 'cannot combine --avoid and --extract-recursive-dependencies\n'
            message += '%s: *** Stop.\n' % APP['name']
            sys.stderr.write(message)
            sys.exit(1)
        modulesystem = GLModuleSystem(config)
        for name in modules:
            module = modulesystem.find(name)
            if module:
                dependencies = module.getDependenciesRecursively()
                dependencies_names = sorted([ m.name
                                              for m in dependencies ])
                sys.stdout.write(lines_to_multiline(dependencies_names))

    elif mode == 'extract-dependents':
        if avoids:
            message = '%s: *** ' % APP['name']
            message += 'cannot combine --avoid and --extract-dependents\n'
            message += '%s: *** Stop.\n' % APP['name']
            sys.stderr.write(message)
            sys.exit(1)
        modulesystem = GLModuleSystem(config)
        for name in modules:
            module = modulesystem.find(name)
            if module:
                dependents = module.getDependents()
                dependents_names = sorted([ m.name
                                            for m in dependents ])
                sys.stdout.write(lines_to_multiline(dependents_names))

    elif mode == 'extract-recursive-dependents':
        if avoids:
            message = '%s: *** ' % APP['name']
            message += 'cannot combine --avoid and --extract-recursive-dependents\n'
            message += '%s: *** Stop.\n' % APP['name']
            sys.stderr.write(message)
            sys.exit(1)
        modulesystem = GLModuleSystem(config)
        for name in modules:
            module = modulesystem.find(name)
            if module:
                dependents = module.getDependentsRecursively()
                dependents_names = sorted([ m.name
                                            for m in dependents ])
                sys.stdout.write(lines_to_multiline(dependents_names))

    elif mode == 'extract-autoconf-snippet':
        modulesystem = GLModuleSystem(config)
        for name in modules:
            module = modulesystem.find(name)
            if module:
                sys.stdout.write(module.getAutoconfSnippet())

    elif mode == 'extract-automake-snippet':
        modulesystem = GLModuleSystem(config)
        for name in modules:
            module = modulesystem.find(name)
            if module:
                sys.stdout.write(module.getAutomakeSnippet())

    elif mode == 'extract-include-directive':
        modulesystem = GLModuleSystem(config)
        for name in modules:
            module = modulesystem.find(name)
            if module:
                sys.stdout.write(module.getInclude())

    elif mode == 'extract-link-directive':
        modulesystem = GLModuleSystem(config)
        for name in modules:
            module = modulesystem.find(name)
            if module:
                sys.stdout.write(module.getLink())

    elif mode == 'extract-recursive-link-directive':
        if avoids:
            message = '%s: *** ' % APP['name']
            message += 'cannot combine --avoid and --extract-recursive-link-directive\n'
            message += '%s: *** Stop.\n' % APP['name']
            sys.stderr.write(message)
            sys.exit(1)
        modulesystem = GLModuleSystem(config)
        for name in modules:
            module = modulesystem.find(name)
            if module:
                sys.stdout.write(module.getLinkDirectiveRecursively())

    elif mode == 'extract-license':
        modulesystem = GLModuleSystem(config)
        for name in modules:
            module = modulesystem.find(name)
            if module:
                print(module.getLicense())

    elif mode == 'extract-maintainer':
        modulesystem = GLModuleSystem(config)
        for name in modules:
            module = modulesystem.find(name)
            if module:
                sys.stdout.write(module.getMaintainer())

    elif mode == 'extract-tests-module':
        modulesystem = GLModuleSystem(config)
        for name in modules:
            module = modulesystem.find(name)
            if module and modulesystem.exists(module.getTestsName()):
                print(module.getTestsName())

    elif mode == 'copy-file':
        srcpath = files[0]
        # The second argument is the destination; either a directory or a file.
        # It defaults to the current directory.
        if len(files) == 2:
            dest = files[1]
        else:  # if len(files) < 2
            dest = '.'
        if not auxdir:
            auxdir = 'build-aux'
        if not sourcebase:
            sourcebase = 'lib'
        if not m4base:
            m4base = 'm4'
        if not docbase:
            docbase = 'doc'
        if not testsbase:
            testsbase = 'tests'
        config.setAuxDir(auxdir)
        config.setSourceBase(sourcebase)
        config.setM4Base(m4base)
        config.setDocBase(docbase)
        config.setTestsBase(testsbase)
        filesystem = GLFileSystem(config)
        lookedup, flag = filesystem.lookup(srcpath)
        if os.path.isdir(dest):
            destdir = dest
            destpath = rewrite_file_name(srcpath, config)
        else:  # if not os.path.isdir(dest)
            destdir = os.path.dirname(dest)
            destpath = os.path.basename(dest)
        # Create the directory for destfile.
        dirname = os.path.dirname(joinpath(destdir, destpath))
        if not config['dryrun']:
            if dirname and not os.path.isdir(dirname):
                try:  # Try to create directories
                    os.makedirs(dirname)
                except FileExistsError:
                    pass
        # Copy the file.
        config.setDestDir(destdir)
        assistant = GLFileAssistant(config)
        tmpfile = assistant.tmpfilename(destpath)
        copyfile(lookedup, tmpfile)
        ensure_writable(tmpfile)
        assistant.setOriginal(srcpath)
        assistant.setRewritten(destpath)
        if os.path.isfile(joinpath(destdir, destpath)):
            # The file already exists.
            assistant.update(lookedup, flag, tmpfile, True)
        else:  # if not os.path.isfile(joinpath(destdir, destpath))
            # Install the file.
            # Don't protest if the file should be there but isn't: it happens
            # frequently that developers don't put autogenerated files under
            # version control.
            assistant.add(lookedup, flag, tmpfile)
        if os.path.isfile(tmpfile):
            os.remove(tmpfile)

    else:
        message = '%s: *** ' % APP['name']
        message += 'no mode specified\n'
        message += '%s: *** Stop.\n' % APP['name']
        sys.stderr.write(message)
        sys.exit(1)

    if copymode == CopyAction.Hardlink or lcopymode == CopyAction.Hardlink:
        # Setting hard links modifies the ctime of files in the gnulib checkout.
        # This disturbs the result of the next "gitk" invocation.
        # Workaround: Let git scan the files. This can be done through
        # "git update-index --refresh" or "git status" or "git diff".
        if os.path.isdir(joinpath(DIRS['root'], '.git')):
            try:
                sp.run(['git', 'update-index', '--refresh'],
                       cwd=DIRS['root'], stdout=sp.DEVNULL)
            except FileNotFoundError:
                # No 'git' program was found.
                pass


def cli_exception(exc_type, exc_value, exc_traceback) -> None:
    '''Exception hook that does not print a traceback for KeyboardInterrupts
    thrown when Ctrl-C is pressed.'''
    if not issubclass(exc_type, KeyboardInterrupt):
        sys.__excepthook__(exc_type, exc_value, exc_traceback)


def main_with_exception_handling() -> None:
    # Don't print tracebacks for KeyboardInterrupts when stdin is a tty.
    if sys.stdin and sys.stdin.isatty():
        sys.excepthook = cli_exception

    try:  # Try to execute
        with tempfile.TemporaryDirectory(prefix='glpy') as temporary_directory:
            main(temporary_directory)
    except KeyboardInterrupt:
        sys.stderr.write('%s: *** Stop.\n' % APP['name'])
        sys.exit(1)
    except GLError as error:
        errmode = 0  # gnulib-style errors
        errno = error.errno
        errinfo = error.errinfo
        if errmode == 0:
            message = '%s: *** ' % APP['name']
            if errinfo == None:
                errinfo = ''
            if errno == 1:
                message += 'file %s not found' % errinfo
            elif errno == 2:
                message += 'patch file %s didn\'t apply cleanly' % errinfo
            elif errno == 3:
                message += 'cannot find %s - make sure you run gnulib-tool from within your package\'s directory' % errinfo
            elif errno == 4:
                message += 'minimum supported autoconf version is 2.64. Try adding'
                message += ' AC_PREREQ([%s])' % DEFAULT_AUTOCONF_MINVERSION
                message += ' to your configure.ac.'
            elif errno == 5:
                message += '%s is expected to contain gl_M4_BASE([%s])' % (repr(os.path.join(errinfo, 'gnulib-comp.m4')), repr(errinfo))
            elif errno == 6:
                message += 'missing --source-base option'
            elif errno == 7:
                message += 'missing --doc-base option. --doc-base has been introduced '
                message += 'on 2006-07-11; if your last invocation of \'gnulib-tool '
                message += '--import\' is before that date, you need to run'
                message += '\'gnulib-tool --import\' once, with a --doc-base option.'
            elif errno == 8:
                message += 'missing --tests-base option'
            elif errno == 9:
                message += 'missing --lib option'
            elif errno == 11:
                incompatibilities = ''
                message += 'incompatible license on modules:\n'
                for pair in errinfo:
                    incompatibilities += pair[0]
                    incompatibilities += ' %s' % pair[1]
                    incompatibilities += '\n'
                tempname = tempfile.mktemp()
                with open(tempname, mode='w', newline='\n', encoding='utf-8') as file:
                    file.write(incompatibilities)
                sed_table = 's,^\\([^ ]*\\) ,\\1' + ' ' * 51 + ',\n'
                sed_table += 's,^\\(' + '.' * 49 + '[^ ]*\\) *,' + ' ' * 17 + '\\1 ,'
                args = ['sed', '-e', sed_table, tempname]
                incompatibilities = sp.check_output(args).decode(ENCS['default'])
                message += incompatibilities
                os.remove(tempname)
            elif errno == 12:
                message += 'refusing to do nothing'
            elif errno == 13:
                message += 'could not create directory %s' % errinfo
            elif errno == 14:
                message += 'could not delete file %s' % errinfo
            elif errno == 15:
                message += 'could not create file %s' % errinfo
            elif errno == 16:
                message += 'could not transform file %s' % errinfo
            elif errno == 17:
                message += 'could not update file %s' % errinfo
            elif errno == 18:
                message += 'module %s lacks a license' % errinfo
            elif errno == 19:
                message += 'could not create destination directory: %s' % errinfo
            elif errno == 20:
                message += 'could not patch test-driver script'
            elif errno == 21:
                message += ('Option --automake-subdir/--automake-subdir-tests are only '
                            'supported if the definition of AUTOMAKE_OPTIONS in '
                            'Makefile.am contains \'subdir-objects\'.')
            elif errno == 22:
                message += 'not overwriting destination directory: %s' % errinfo
            elif errno == 23:
                message += "module %s doesn't exist" % errinfo
            elif errno == 24:
                message += 'module %s cannot be used in a testdir' % errinfo
            message += '\n%s: *** Stop.\n' % APP['name']
            sys.stderr.write(message)
            sys.exit(1)



# Copyright (C) 2002-2026 Free Software Foundation, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

#===============================================================================
# Define global imports
#===============================================================================
from enum import Enum


#===============================================================================
# Define CopyAction class
#===============================================================================
class CopyAction(Enum):
    Copy = 0
    Symlink = 1
    Hardlink = 2



# Copyright (C) 2002-2026 Free Software Foundation, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

'''Gnulib - The GNU Portability Library
GNU software has a well-deserved reputation for running on many different types
of systems. While our primary goal is to write software for the GNU system, many
users and developers have been introduced to us through the systems that they
were already using.
Gnulib is a central location for common GNU code, intended to be shared among
GNU packages. GCC has libiberty, but this is hard to disentangle from the GCC
build tree.
Gnulib takes a different approach. Its components are intended to be shared at
the source level, rather than being a library that gets built, installed, and
linked against. Thus, there is no distribution tarball; the idea is to copy
files from Gnulib into your own source tree.
Gnulib also includes copies of a few files purely for convenience: the GNU
coding standards, the GNU maintainer information, the GPL and other licenses (in
Texinfo), assorted configuration scripts, and more. The goal is to provide all
the common infrastructure needed by GNU packages.'''

__copyright__ = '2002-2026 Free Software Foundation, Inc.'
__author__ = \
    [
        'Bruno Haible',
        'Paul Eggert',
        'Simon Josefsson',
        'Dmitry Selyutin',
        'Collin Funk',
    ]
__license__ = 'GNU GPLv3+'
